from django.contrib import messages
from django.db import transaction
from django.utils.translation import gettext_lazy as _

from ....admin.views import generic
from ...models import Attachment, Post
from ..forms import FilterAttachmentsForm


class AttachmentAdmin(generic.AdminBaseMixin):
    root_link = "misago:admin:attachments:index"
    model = Attachment
    templates_dir = "misago/admin/attachments"
    message_404 = _("Requested attachment could not be found.")

    def get_queryset(self):
        qs = super().get_queryset()
        return qs.select_related(
            "filetype", "uploader", "post", "post__thread", "post__category"
        )


class AttachmentsList(AttachmentAdmin, generic.ListView):
    items_per_page = 20
    ordering = [
        ("-id", _("From newest")),
        ("id", _("From oldest")),
        ("filename", _("A to z")),
        ("-filename", _("Z to a")),
        ("size", _("Smallest files")),
        ("-size", _("Largest files")),
    ]
    selection_label = _("With attachments: 0")
    empty_selection_label = _("Select attachments")
    mass_actions = [
        {
            "action": "delete",
            "name": _("Delete attachments"),
            "confirmation": _("Are you sure you want to delete selected attachments?"),
            "is_atomic": False,
        }
    ]
    filter_form = FilterAttachmentsForm

    def action_delete(self, request, attachments):
        deleted_attachments = []
        desynced_posts = []

        for attachment in attachments:
            if attachment.post:
                deleted_attachments.append(attachment.pk)
                desynced_posts.append(attachment.post_id)

        if desynced_posts:
            with transaction.atomic():
                for post in Post.objects.filter(id__in=desynced_posts):
                    self.delete_from_cache(post, deleted_attachments)

        for attachment in attachments:
            attachment.delete()

        message = _("Selected attachments have been deleted.")
        messages.success(request, message)

    def delete_from_cache(self, post, attachments):
        if not post.attachments_cache:
            return  # admin action may be taken due to desynced state

        clean_cache = []
        for a in post.attachments_cache:
            if a["id"] not in attachments:
                clean_cache.append(a)

        post.attachments_cache = clean_cache or None
        post.save(update_fields=["attachments_cache"])


class DeleteAttachment(AttachmentAdmin, generic.ButtonView):
    def button_action(self, request, target):
        if target.post:
            self.delete_from_cache(target)
        target.delete()
        message = _('Attachment "%(filename)s" has been deleted.')
        messages.success(request, message % {"filename": target.filename})

    def delete_from_cache(self, attachment):
        if not attachment.post.attachments_cache:
            return  # admin action may be taken due to desynced state

        clean_cache = []
        for a in attachment.post.attachments_cache:
            if a["id"] != attachment.id:
                clean_cache.append(a)

        attachment.post.attachments_cache = clean_cache or None
        attachment.post.save(update_fields=["attachments_cache"])



from django.urls import reverse

from ... import test
from ....admin.test import AdminTestCase
from ....categories.models import Category
from ...models import Attachment, AttachmentType


class AttachmentAdminViewsTests(AdminTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")
        self.post = test.post_thread(category=self.category).first_post

        self.filetype = AttachmentType.objects.order_by("id").first()

        self.admin_link = reverse("misago:admin:attachments:index")

    def mock_attachment(self, post=None, file=None, image=None, thumbnail=None):
        return Attachment.objects.create(
            secret=Attachment.generate_new_secret(),
            filetype=self.filetype,
            post=post,
            size=1000,
            uploader=self.user,
            uploader_name=self.user.username,
            uploader_slug=self.user.slug,
            filename="testfile_%s.zip" % (Attachment.objects.count() + 1),
            file=None,
            image=None,
            thumbnail=None,
        )

    def test_link_registered(self):
        """admin nav contains attachments link"""
        response = self.client.get(reverse("misago:admin:settings:index"))
        self.assertContains(response, self.admin_link)

    def test_list_view(self):
        """attachments list returns 200 and renders all attachments"""
        final_link = self.client.get(self.admin_link)["location"]

        response = self.client.get(final_link)
        self.assertEqual(response.status_code, 200)

        attachments = [
            self.mock_attachment(self.post, file="somefile.pdf"),
            self.mock_attachment(image="someimage.jpg"),
            self.mock_attachment(
                self.post, image="somelargeimage.png", thumbnail="somethumb.png"
            ),
        ]

        response = self.client.get(final_link)
        self.assertEqual(response.status_code, 200)

        for attachment in attachments:
            delete_link = reverse(
                "misago:admin:attachments:delete", kwargs={"pk": attachment.pk}
            )
            self.assertContains(response, attachment.filename)
            self.assertContains(response, delete_link)
            self.assertContains(response, attachment.get_absolute_url())
            self.assertContains(response, attachment.uploader.username)
            self.assertContains(response, attachment.uploader.get_absolute_url())

            if attachment.thumbnail:
                self.assertContains(response, attachment.get_thumbnail_url())

    def test_delete_multiple(self):
        """mass delete tool on list works"""
        attachments = [
            self.mock_attachment(self.post, file="somefile.pdf"),
            self.mock_attachment(image="someimage.jpg"),
            self.mock_attachment(
                self.post, image="somelargeimage.png", thumbnail="somethumb.png"
            ),
        ]

        self.post.attachments_cache = [{"id": attachments[-1].pk}]
        self.post.save()

        response = self.client.post(
            self.admin_link,
            data={"action": "delete", "selected_items": [a.pk for a in attachments]},
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Attachment.objects.count(), 0)

        # assert attachments were removed from post's cache
        attachments_cache = self.category.post_set.get(
            pk=self.post.pk
        ).attachments_cache
        self.assertIsNone(attachments_cache)

    def test_delete_view(self):
        """delete attachment view has no showstoppers"""
        attachment = self.mock_attachment(self.post)
        self.post.attachments_cache = [
            {"id": attachment.pk + 1},
            {"id": attachment.pk},
            {"id": attachment.pk + 2},
        ]
        self.post.save()

        action_link = reverse(
            "misago:admin:attachments:delete", kwargs={"pk": attachment.pk}
        )

        response = self.client.post(action_link)
        self.assertEqual(response.status_code, 302)

        # clean alert about item, grab final list url
        final_link = self.client.get(self.admin_link)["location"]

        response = self.client.get(final_link)
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, action_link)

        # assert it was removed from post's attachments cache
        attachments_cache = self.category.post_set.get(
            pk=self.post.pk
        ).attachments_cache
        self.assertEqual(
            attachments_cache, [{"id": attachment.pk + 1}, {"id": attachment.pk + 2}]
        )



from django.core import mail
from django.urls import reverse

from ...conf.test import override_dynamic_settings
from ..test import AuthenticatedUserTestCase


class OptionsViewsTests(AuthenticatedUserTestCase):
    def test_lander_view_returns_200(self):
        """/options has no show stoppers"""
        response = self.client.get(reverse("misago:options"))
        self.assertEqual(response.status_code, 200)

    def test_form_view_returns_200(self):
        """/options/some-form has no show stoppers"""
        response = self.client.get(
            reverse("misago:options-form", kwargs={"form_name": "some-fake-form"})
        )
        self.assertEqual(response.status_code, 200)


class ConfirmChangeEmailTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()
        link = "/api/users/%s/change-email/" % self.user.pk

        with override_dynamic_settings(forum_address="http://test.com/"):
            response = self.client.post(
                link,
                data={"new_email": "n3w@email.com", "password": self.USER_PASSWORD},
            )
            self.assertEqual(response.status_code, 200)

        for line in [l.strip() for l in mail.outbox[0].body.splitlines()]:
            if line.startswith("http://"):
                self.link = line.strip()
                break

    def test_invalid_token(self):
        """invalid token is rejected"""
        response = self.client.get(
            reverse("misago:options-confirm-email-change", kwargs={"token": "invalid"})
        )

        self.assertContains(
            response, "Change confirmation link is invalid.", status_code=400
        )

    def test_change_email(self):
        """valid token changes email"""
        response = self.client.get(self.link)

        self.assertContains(response, "your e-mail has been changed")

        self.reload_user()
        self.assertEqual(self.user.email, "n3w@email.com")

    @override_dynamic_settings(
        enable_oauth2_client=True,
        oauth2_provider="Lorem",
    )
    def test_change_email_view_returns_403_if_oauth_is_enabled(self):
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 403)


class ConfirmChangePasswordTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()
        link = "/api/users/%s/change-password/" % self.user.pk

        with override_dynamic_settings(forum_address="http://test.com/"):
            response = self.client.post(
                link,
                data={"new_password": "n3wp4ssword", "password": self.USER_PASSWORD},
            )
            self.assertEqual(response.status_code, 200)

        for line in [l.strip() for l in mail.outbox[0].body.splitlines()]:
            if line.startswith("http://"):
                self.link = line.strip()
                break

    def test_invalid_token(self):
        """invalid token is rejected"""
        response = self.client.get(
            reverse(
                "misago:options-confirm-password-change", kwargs={"token": "invalid"}
            )
        )

        self.assertContains(
            response, "Change confirmation link is invalid.", status_code=400
        )

    def test_change_password(self):
        """valid token changes password"""
        response = self.client.get(self.link)

        self.assertContains(response, "your password has been changed")

        self.reload_user()
        self.assertFalse(self.user.check_password(self.USER_PASSWORD))
        self.assertTrue(self.user.check_password("n3wp4ssword"))

    @override_dynamic_settings(
        enable_oauth2_client=True,
        oauth2_provider="Lorem",
    )
    def test_change_password_view_returns_403_if_oauth_is_enabled(self):
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 403)



import pytest
from django.test import override_settings


def test_accessing_attr_returns_setting_value_defined_in_settings_file(settings):
    assert settings.STATIC_URL


def test_accessing_attr_returns_setting_value_defined_in_misago_defaults_file(settings):
    assert settings.MISAGO_MOMENT_JS_LOCALES


def test_setting_value_can_be_overridden_using_django_util(settings):
    with override_settings(STATIC_URL="/test/"):
        assert settings.STATIC_URL == "/test/"


def test_default_setting_value_can_be_overridden_using_django_util(settings):
    with override_settings(MISAGO_MOMENT_JS_LOCALES="test"):
        assert settings.MISAGO_MOMENT_JS_LOCALES == "test"


def test_undefined_setting_value_can_be_overridden_using_django_util(settings):
    with override_settings(UNDEFINED_SETTING="test"):
        assert settings.UNDEFINED_SETTING == "test"


def test_accessing_attr_for_undefined_setting_raises_attribute_error(settings):
    with pytest.raises(AttributeError):
        assert settings.UNDEFINED_SETTING



from ...users import tokens


def test_token_can_be_created_for_user(user):
    assert tokens.make(user, "test")


def test_token_can_be_validated(user):
    token = tokens.make(user, "test")
    assert tokens.is_valid(user, "test", token)


def test_token_fails_validation_for_different_type(user):
    token = tokens.make(user, "activation")
    assert not tokens.is_valid(user, "new_password", token)


def test_token_fails_validation_for_different_user(user, other_user):
    token = tokens.make(user, "test")
    assert not tokens.is_valid(other_user, "test", token)



import json
from datetime import timedelta

from django.urls import reverse
from django.utils import timezone

from .. import test
from ..models import Post, Thread
from ..test import patch_category_acl
from .test_threads_api import ThreadsApiTestCase


class PostBulkDeleteApiTests(ThreadsApiTestCase):
    def setUp(self):
        super().setUp()

        self.posts = [
            test.reply_thread(self.thread, poster=self.user),
            test.reply_thread(self.thread),
            test.reply_thread(self.thread, poster=self.user),
        ]

        self.api_link = reverse(
            "misago:api:thread-post-list", kwargs={"thread_pk": self.thread.pk}
        )

    def delete(self, url, data=None):
        return self.client.delete(
            url, json.dumps(data), content_type="application/json"
        )

    def test_delete_anonymous(self):
        """api validates if deleting user is authenticated"""
        self.logout_user()

        response = self.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This action is not available to guests."}
        )

    def test_delete_no_data(self):
        """api handles empty data"""
        response = self.client.delete(self.api_link, content_type="application/json")
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": 'Expected a list of items but got type "dict".'}
        )

    def test_delete_no_ids(self):
        """api requires ids to delete"""
        response = self.delete(self.api_link)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "You have to specify at least one post to delete."},
        )

    def test_delete_empty_ids(self):
        """api requires ids to delete"""
        response = self.delete(self.api_link, [])
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "You have to specify at least one post to delete."},
        )

    @patch_category_acl({"can_hide_posts": 2})
    def test_validate_ids(self):
        """api validates that ids are list of ints"""
        response = self.delete(self.api_link, True)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": 'Expected a list of items but got type "bool".'}
        )

        response = self.delete(self.api_link, "abbss")
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": 'Expected a list of items but got type "str".'}
        )

        response = self.delete(self.api_link, [1, 2, 3, "a", "b", "x"])
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "One or more post ids received were invalid."}
        )

    @patch_category_acl({"can_hide_posts": 2})
    def test_validate_ids_length(self):
        """api validates that ids are list of ints"""
        response = self.delete(self.api_link, list(range(100)))
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "No more than 24 posts can be deleted at a single time."},
        )

    @patch_category_acl({"can_hide_posts": 2})
    def test_validate_posts_exist(self):
        """api validates that ids are visible posts"""
        response = self.delete(self.api_link, [p.id * 10 for p in self.posts])
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "One or more posts to delete could not be found."},
        )

    @patch_category_acl({"can_hide_posts": 2})
    def test_validate_posts_visibility(self):
        """api validates that ids are visible posts"""
        self.posts[1].is_unapproved = True
        self.posts[1].save()

        response = self.delete(self.api_link, [p.id for p in self.posts])
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "One or more posts to delete could not be found."},
        )

    @patch_category_acl({"can_hide_posts": 2})
    def test_validate_posts_same_thread(self):
        """api validates that ids are same thread posts"""
        other_thread = test.post_thread(category=self.category)
        self.posts.append(test.reply_thread(other_thread, poster=self.user))

        response = self.delete(self.api_link, [p.id for p in self.posts])
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "One or more posts to delete could not be found."},
        )

    @patch_category_acl({"can_hide_posts": 1, "can_hide_own_posts": 1})
    def test_no_permission(self):
        """api validates permission to delete"""
        response = self.delete(self.api_link, [p.id for p in self.posts])
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't delete posts in this category."}
        )

    @patch_category_acl(
        {"can_hide_posts": 0, "can_hide_own_posts": 2, "post_edit_time": 10}
    )
    def test_delete_other_user_post_no_permission(self):
        """api valdiates if user can delete other users posts"""
        response = self.delete(self.api_link, [p.id for p in self.posts])
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't delete other users posts in this category."},
        )

    @patch_category_acl(
        {"can_hide_posts": 0, "can_hide_own_posts": 2, "can_protect_posts": False}
    )
    def test_delete_protected_post_no_permission(self):
        """api validates if user can delete protected post"""
        self.posts[0].is_protected = True
        self.posts[0].save()

        response = self.delete(self.api_link, [p.id for p in self.posts])
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This post is protected. You can't delete it."}
        )

    @patch_category_acl(
        {"can_hide_posts": 0, "can_hide_own_posts": 2, "post_edit_time": 1}
    )
    def test_delete_protected_post_after_edit_time(self):
        """api validates if user can delete delete post after edit time"""
        self.posts[0].posted_on = timezone.now() - timedelta(minutes=10)
        self.posts[0].save()

        response = self.delete(self.api_link, [p.id for p in self.posts])
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't delete posts that are older than 1 minute."},
        )

    @patch_category_acl(
        {"can_hide_posts": 2, "can_hide_own_posts": 2, "can_close_threads": False}
    )
    def test_delete_post_closed_thread_no_permission(self):
        """api valdiates if user can delete posts in closed threads"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.delete(self.api_link, [p.id for p in self.posts])
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This thread is closed. You can't delete posts in it."},
        )

    @patch_category_acl(
        {"can_hide_posts": 2, "can_hide_own_posts": 2, "can_close_threads": False}
    )
    def test_delete_post_closed_category_no_permission(self):
        """api valdiates if user can delete posts in closed categories"""
        self.category.is_closed = True
        self.category.save()

        response = self.delete(self.api_link, [p.id for p in self.posts])
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This category is closed. You can't delete posts in it."},
        )

    @patch_category_acl({"can_hide_posts": 2, "can_hide_own_posts": 2})
    def test_delete_first_post(self):
        """api disallows first post's deletion"""
        ids = [p.id for p in self.posts]
        ids.append(self.thread.first_post_id)

        response = self.delete(self.api_link, ids)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't delete thread's first post."}
        )

    @patch_category_acl({"can_hide_posts": 2, "can_hide_own_posts": 2})
    def test_delete_best_answer(self):
        """api disallows best answer deletion"""
        self.thread.set_best_answer(self.user, self.posts[0])
        self.thread.save()

        response = self.delete(self.api_link, [p.id for p in self.posts])
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't delete this post because its marked as best answer."},
        )

    @patch_category_acl(
        {"can_hide_posts": 2, "can_hide_own_posts": 2, "can_hide_events": 0}
    )
    def test_delete_event(self):
        """api differs posts from events"""
        self.posts[1].is_event = True
        self.posts[1].save()

        response = self.delete(self.api_link, [p.id for p in self.posts])
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't delete events in this category."}
        )

    @patch_category_acl(
        {"can_hide_posts": 0, "can_hide_own_posts": 2, "post_edit_time": 10}
    )
    def test_delete_owned_posts(self):
        """api deletes owned thread posts"""
        ids = [self.posts[0].id, self.posts[-1].id]

        response = self.delete(self.api_link, ids)
        self.assertEqual(response.status_code, 200)

        self.thread = Thread.objects.get(pk=self.thread.pk)
        self.assertNotEqual(self.thread.last_post_id, ids[-1])

        for post in ids:
            with self.assertRaises(Post.DoesNotExist):
                self.thread.post_set.get(pk=post)

    @patch_category_acl({"can_hide_posts": 2, "can_hide_own_posts": 0})
    def test_delete_posts(self):
        """api deletes thread posts"""
        response = self.delete(self.api_link, [p.id for p in self.posts])
        self.assertEqual(response.status_code, 200)

        self.thread = Thread.objects.get(pk=self.thread.pk)
        self.assertNotEqual(self.thread.last_post_id, self.posts[-1].pk)

        for post in self.posts:
            with self.assertRaises(Post.DoesNotExist):
                self.thread.post_set.get(pk=post.pk)



from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _


class DeletedUser(models.Model):
    DELETED_BY_SELF = 1
    DELETED_BY_STAFF = 2
    DELETED_BY_SYSTEM = 3

    DELETED_BY_CHOICES = (
        (DELETED_BY_SELF, _("By self")),
        (DELETED_BY_STAFF, _("By staff")),
        (DELETED_BY_SYSTEM, _("By system")),
    )

    deleted_on = models.DateTimeField(default=timezone.now, db_index=True)
    deleted_by = models.PositiveIntegerField(choices=DELETED_BY_CHOICES, db_index=True)

    class Meta:
        ordering = ["-id"]



from ..cache.versions import get_cache_versions
from .dynamicsettings import DynamicSettings


def get_dynamic_settings():
    cache_versions = get_cache_versions()
    return DynamicSettings(cache_versions)



from urllib.parse import urlencode

from django.urls import reverse

list_link = reverse("misago:admin:users:index")


def test_view_redirects_if_redirected_flag_is_not_present_in_querystring(admin_client):
    response = admin_client.get(list_link)
    assert response.status_code == 302


def test_view_sets_redirect_flag_in_redirect_url(admin_client):
    response = admin_client.get(list_link)
    assert response.status_code == 302
    assert "redirected=1" in response["location"]


def test_view_checks_only_redirect_flag_presence_and_not_value(admin_client):
    response = admin_client.get(list_link + "?redirected")
    assert response.status_code == 200


def test_view_preserves_rest_of_querystring_in_redirect_url(admin_client):
    response = admin_client.get(list_link + "?username=test")
    assert response.status_code == 302
    assert "redirected=1" in response["location"]
    assert "username=test" in response["location"]


def test_unicode_is_preserved_in_redirect_querystring(admin_client):
    response = admin_client.get(list_link + "?username=łóć")
    assert response.status_code == 302
    assert urlencode({"username": "łóć"}) in response["location"]


def test_view_is_not_redirecting_if_flag_is_set_in_querystring(admin_client):
    response = admin_client.get("%s?redirected=1" % list_link)
    assert response.status_code == 200


def test_restoring_filters_from_session_handles_filters_entry_being_none(admin_client):
    """Regression test for https://github.com/rafalp/Misago/pull/1177"""
    response = admin_client.get("%s?set_filters=1&rank=1" % list_link)
    assert response.status_code == 302
    response = admin_client.get("%s?set_filters=0" % list_link)
    assert response.status_code == 302
    response = admin_client.get("%s?redirected=1" % list_link)
    assert response.status_code == 200



import os
from hashlib import md5
from io import BytesIO

from django.core.files.base import ContentFile
from django.utils.crypto import get_random_string
from PIL.Image import Resampling

from ...conf import settings


def normalize_image(image):
    """strip image of animation, convert to RGBA"""
    image.seek(0)
    return image.copy().convert("RGBA")


def delete_avatar(user, delete_tmp=True, delete_src=True):
    if delete_tmp and user.avatar_tmp:
        user.avatar_tmp.delete(save=False)

    if delete_src and user.avatar_src:
        user.avatar_src.delete(save=False)

    for avatar in user.avatar_set.all():
        avatar.image.delete(save=False)
    user.avatar_set.all().delete()


def store_avatar(user, image):
    from ..models import Avatar

    image = normalize_image(image)

    avatars = []
    for size in sorted(settings.MISAGO_AVATARS_SIZES, reverse=True):
        image_stream = BytesIO()

        image = image.resize((size, size), Resampling.LANCZOS)
        image.save(image_stream, "PNG")

        avatars.append(
            Avatar.objects.create(
                user=user,
                size=size,
                image=ContentFile(image_stream.getvalue(), "avatar"),
            )
        )

    user.avatars = [{"size": a.size, "url": a.url} for a in avatars]
    user.save(update_fields=["avatars"])


def store_new_avatar(user, image, delete_tmp=True, delete_src=True):
    delete_avatar(user, delete_tmp=delete_tmp, delete_src=delete_src)
    store_avatar(user, image)


def store_temporary_avatar(user, image):
    image_stream = BytesIO()

    normalize_image(image)
    image.save(image_stream, "PNG")

    if user.avatar_tmp:
        user.avatar_tmp.delete(save=False)

    user.avatar_tmp = ContentFile(image_stream.getvalue(), "avatar")
    user.save(update_fields=["avatar_tmp"])


def store_original_avatar(user):
    if user.avatar_src:
        user.avatar_src.delete(save=False)
    user.avatar_src = user.avatar_tmp
    user.avatar_tmp = None
    user.save(update_fields=["avatar_tmp", "avatar_src"])


def upload_to(instance, filename):
    spread_path = md5(get_random_string(64).encode()).hexdigest()
    secret = get_random_string(32)
    filename_clean = "%s.png" % get_random_string(32)

    return os.path.join(
        "avatars", spread_path[:2], spread_path[2:4], secret, filename_clean
    )



from django.core import mail
from django.test import TestCase
from django.urls import reverse

from ...cache.versions import get_cache_versions
from ...conf.dynamicsettings import DynamicSettings
from ...conf.test import override_dynamic_settings
from ...users.test import create_test_user
from ..mail import build_mail, mail_user, mail_users


class MailTests(TestCase):
    def test_building_mail_without_context_raises_value_error(self):
        user = create_test_user("User", "user@example.com")
        with self.assertRaises(ValueError):
            build_mail(user, "Misago Test Mail", "misago/emails/base")

    def test_building_mail_without_settings_in_context_raises_value_error(self):
        user = create_test_user("User", "user@example.com")
        with self.assertRaises(ValueError):
            build_mail(
                user, "Misago Test Mail", "misago/emails/base", context={"settings": {}}
            )

    @override_dynamic_settings(forum_address="http://test.com/")
    def test_mail_user(self):
        """mail_user sets message in backend"""
        user = create_test_user("User", "user@example.com")

        cache_versions = get_cache_versions()
        settings = DynamicSettings(cache_versions)

        mail_user(
            user,
            "Misago Test Mail",
            "misago/emails/base",
            context={"settings": settings},
        )

        self.assertEqual(mail.outbox[0].subject, "Misago Test Mail")

        # assert that url to user's avatar is valid
        html_body = mail.outbox[0].alternatives[0][0]
        user_avatar_url = reverse(
            "misago:user-avatar", kwargs={"pk": user.pk, "size": 32}
        )

        self.assertIn(user_avatar_url, html_body)

    def test_mail_users(self):
        """mail_users sets messages in backend"""
        cache_versions = get_cache_versions()
        settings = DynamicSettings(cache_versions)

        test_users = [
            create_test_user("User1", "User1@example.com"),
            create_test_user("Use2r", "User2@example.com"),
            create_test_user("Us3er", "User3@example.com"),
            create_test_user("U4ser", "User4@example.com"),
            create_test_user("5User", "User5@example.com"),
        ]

        mail_users(
            test_users,
            "Misago Test Spam",
            "misago/emails/base",
            context={"settings": settings},
        )

        spams_sent = 0
        for message in mail.outbox:
            if message.subject == "Misago Test Spam":
                spams_sent += 1

        self.assertEqual(spams_sent, len(test_users))



from django.core.exceptions import PermissionDenied, ValidationError
from django.http import Http404
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy, ngettext
from rest_framework import serializers

from ...acl.objectacl import add_acl_to_obj
from ...categories import THREADS_ROOT_NAME
from ...conf import settings
from ..mergeconflict import MergeConflict
from ..models import Thread
from ..permissions import (
    allow_delete_best_answer,
    allow_delete_event,
    allow_delete_post,
    allow_delete_thread,
    allow_merge_post,
    allow_merge_thread,
    allow_move_post,
    allow_split_post,
    can_reply_thread,
    can_see_thread,
    can_start_thread,
    exclude_invisible_posts,
)
from ..threadtypes import trees_map
from ..utils import get_thread_id_from_url
from ..validators import validate_category, validate_thread_title


__all__ = [
    "DeletePostsSerializer",
    "DeleteThreadsSerializer",
    "MergePostsSerializer",
    "MergeThreadSerializer",
    "MergeThreadsSerializer",
    "MovePostsSerializer",
    "NewThreadSerializer",
    "SplitPostsSerializer",
]


def get_posts_limit(settings):
    return settings.posts_per_page + settings.posts_per_page_orphans


class DeletePostsSerializer(serializers.Serializer):
    error_empty_or_required = gettext_lazy(
        "You have to specify at least one post to delete."
    )

    posts = serializers.ListField(
        allow_empty=False,
        child=serializers.IntegerField(
            error_messages={
                "invalid": gettext_lazy("One or more post ids received were invalid.")
            }
        ),
        error_messages={
            "required": error_empty_or_required,
            "null": error_empty_or_required,
            "empty": error_empty_or_required,
        },
    )

    def validate_posts(self, data):
        limit = get_posts_limit(self.context["settings"])
        if len(data) > limit:
            message = ngettext(
                "No more than %(limit)s post can be deleted at a single time.",
                "No more than %(limit)s posts can be deleted at a single time.",
                limit,
            )
            raise ValidationError(message % {"limit": limit})

        user_acl = self.context["user_acl"]
        thread = self.context["thread"]

        posts_queryset = exclude_invisible_posts(
            user_acl, thread.category, thread.post_set
        )
        posts_queryset = posts_queryset.filter(id__in=data).order_by("id")

        posts = []
        for post in posts_queryset:
            post.category = thread.category
            post.thread = thread

            if post.is_event:
                allow_delete_event(user_acl, post)
            else:
                allow_delete_best_answer(user_acl, post)
                allow_delete_post(user_acl, post)

            posts.append(post)

        if len(posts) != len(data):
            raise PermissionDenied(_("One or more posts to delete could not be found."))

        return posts


class MergePostsSerializer(serializers.Serializer):
    error_empty_or_required = gettext_lazy(
        "You have to select at least two posts to merge."
    )

    posts = serializers.ListField(
        child=serializers.IntegerField(
            error_messages={
                "invalid": gettext_lazy("One or more post ids received were invalid.")
            }
        ),
        error_messages={
            "null": error_empty_or_required,
            "required": error_empty_or_required,
        },
    )

    def validate_posts(self, data):
        limit = get_posts_limit(self.context["settings"])
        data = list(set(data))

        if len(data) < 2:
            raise serializers.ValidationError(self.error_empty_or_required)
        if len(data) > limit:
            message = ngettext(
                "No more than %(limit)s post can be merged at a single time.",
                "No more than %(limit)s posts can be merged at a single time.",
                limit,
            )
            raise serializers.ValidationError(message % {"limit": limit})

        user_acl = self.context["user_acl"]
        thread = self.context["thread"]

        posts_queryset = exclude_invisible_posts(
            user_acl, thread.category, thread.post_set
        )
        posts_queryset = posts_queryset.filter(id__in=data).order_by("id")

        posts = []
        for post in posts_queryset:
            post.category = thread.category
            post.thread = thread

            try:
                allow_merge_post(user_acl, post)
            except PermissionDenied as e:
                raise serializers.ValidationError(e)

            if not posts:
                posts.append(post)
                continue

            authorship_error = _("Posts made by different users can't be merged.")
            if post.poster_id != posts[0].poster_id:
                raise serializers.ValidationError(authorship_error)
            elif (
                post.poster_id is None
                and posts[0].poster_id is None
                and post.poster_name != posts[0].poster_name
            ):
                raise serializers.ValidationError(authorship_error)

            if posts[0].is_first_post and post.is_best_answer:
                raise serializers.ValidationError(
                    _(
                        "Post marked as best answer can't be merged with "
                        "thread's first post."
                    )
                )

            if not posts[0].is_first_post:
                if (
                    posts[0].is_hidden != post.is_hidden
                    or posts[0].is_unapproved != post.is_unapproved
                ):
                    raise serializers.ValidationError(
                        _("Posts with different visibility can't be merged.")
                    )

            posts.append(post)

        if len(posts) != len(data):
            raise serializers.ValidationError(
                _("One or more posts to merge could not be found.")
            )

        return posts


class MovePostsSerializer(serializers.Serializer):
    error_empty_or_required = gettext_lazy(
        "You have to specify at least one post to move."
    )

    new_thread = serializers.CharField(
        error_messages={"required": gettext_lazy("Enter link to new thread.")}
    )
    posts = serializers.ListField(
        allow_empty=False,
        child=serializers.IntegerField(
            error_messages={
                "invalid": gettext_lazy("One or more post ids received were invalid.")
            }
        ),
        error_messages={
            "empty": error_empty_or_required,
            "null": error_empty_or_required,
            "required": error_empty_or_required,
        },
    )

    def validate_new_thread(self, data):
        request = self.context["request"]
        thread = self.context["thread"]
        viewmodel = self.context["viewmodel"]

        new_thread_id = get_thread_id_from_url(request, data)
        if not new_thread_id:
            raise serializers.ValidationError(_("This is not a valid thread link."))
        if new_thread_id == thread.pk:
            raise serializers.ValidationError(
                _("Thread to move posts to is same as current one.")
            )

        try:
            new_thread = viewmodel(request, new_thread_id).unwrap()
        except Http404:
            raise serializers.ValidationError(
                _(
                    "The thread you have entered link to doesn't "
                    "exist or you don't have permission to see it."
                )
            )

        if not new_thread.acl["can_reply"]:
            raise serializers.ValidationError(
                _("You can't move posts to threads you can't reply.")
            )

        return new_thread

    def validate_posts(self, data):
        limit = get_posts_limit(self.context["settings"])
        data = list(set(data))
        if len(data) > limit:
            message = ngettext(
                "No more than %(limit)s post can be moved at a single time.",
                "No more than %(limit)s posts can be moved at a single time.",
                limit,
            )
            raise serializers.ValidationError(message % {"limit": limit})

        request = self.context["request"]
        thread = self.context["thread"]

        posts_queryset = exclude_invisible_posts(
            request.user_acl, thread.category, thread.post_set
        )
        posts_queryset = posts_queryset.filter(id__in=data).order_by("id")

        posts = []
        for post in posts_queryset:
            post.category = thread.category
            post.thread = thread

            try:
                allow_move_post(request.user_acl, post)
                posts.append(post)
            except PermissionDenied as e:
                raise serializers.ValidationError(e)

        if len(posts) != len(data):
            raise serializers.ValidationError(
                _("One or more posts to move could not be found.")
            )

        return posts


class NewThreadSerializer(serializers.Serializer):
    title = serializers.CharField()
    category = serializers.IntegerField()
    weight = serializers.IntegerField(
        required=False,
        allow_null=True,
        max_value=Thread.WEIGHT_GLOBAL,
        min_value=Thread.WEIGHT_DEFAULT,
    )
    is_hidden = serializers.BooleanField(required=False, allow_null=True)
    is_closed = serializers.BooleanField(required=False, allow_null=True)

    def validate_title(self, title):
        settings = self.context["settings"]
        validate_thread_title(settings, title)
        return title

    def validate_category(self, category_id):
        user_acl = self.context["user_acl"]
        self.category = validate_category(user_acl, category_id)
        if not can_start_thread(user_acl, self.category):
            raise ValidationError(
                _("You can't create new threads in selected category.")
            )
        return self.category

    def validate_weight(self, weight):
        try:
            add_acl_to_obj(self.context["user_acl"], self.category)
        except AttributeError:
            return weight  # don't validate weight further if category failed

        if weight > self.category.acl.get("can_pin_threads", 0):
            if weight == 2:
                raise ValidationError(
                    _(
                        "You don't have permission to pin threads globally "
                        "in this category."
                    )
                )
            else:
                raise ValidationError(
                    _("You don't have permission to pin threads in this category.")
                )
        return weight

    def validate_is_hidden(self, is_hidden):
        try:
            add_acl_to_obj(self.context["user_acl"], self.category)
        except AttributeError:
            return is_hidden  # don't validate hidden further if category failed

        if is_hidden and not self.category.acl.get("can_hide_threads"):
            raise ValidationError(
                _("You don't have permission to hide threads in this category.")
            )
        return is_hidden

    def validate_is_closed(self, is_closed):
        try:
            add_acl_to_obj(self.context["user_acl"], self.category)
        except AttributeError:
            return is_closed  # don't validate closed further if category failed

        if is_closed and not self.category.acl.get("can_close_threads"):
            raise ValidationError(
                _("You don't have permission to close threads in this category.")
            )
        return is_closed


class SplitPostsSerializer(NewThreadSerializer):
    error_empty_or_required = gettext_lazy(
        "You have to specify at least one post to split."
    )

    posts = serializers.ListField(
        allow_empty=False,
        child=serializers.IntegerField(
            error_messages={
                "invalid": gettext_lazy("One or more post ids received were invalid.")
            }
        ),
        error_messages={
            "empty": error_empty_or_required,
            "null": error_empty_or_required,
            "required": error_empty_or_required,
        },
    )

    def validate_posts(self, data):
        limit = get_posts_limit(self.context["settings"])
        if len(data) > limit:
            message = ngettext(
                "No more than %(limit)s post can be split at a single time.",
                "No more than %(limit)s posts can be split at a single time.",
                limit,
            )
            raise ValidationError(message % {"limit": limit})

        thread = self.context["thread"]
        user_acl = self.context["user_acl"]

        posts_queryset = exclude_invisible_posts(
            user_acl, thread.category, thread.post_set
        )
        posts_queryset = posts_queryset.filter(id__in=data).order_by("id")

        posts = []
        for post in posts_queryset:
            post.category = thread.category
            post.thread = thread

            try:
                allow_split_post(user_acl, post)
            except PermissionDenied as e:
                raise ValidationError(str(e))

            posts.append(post)

        if len(posts) != len(data):
            raise ValidationError(_("One or more posts to split could not be found."))

        return posts


class DeleteThreadsSerializer(serializers.Serializer):
    error_empty_or_required = gettext_lazy(
        "You have to specify at least one thread to delete."
    )

    threads = serializers.ListField(
        allow_empty=False,
        child=serializers.IntegerField(
            error_messages={
                "invalid": gettext_lazy("One or more thread ids received were invalid.")
            }
        ),
        error_messages={
            "required": error_empty_or_required,
            "null": error_empty_or_required,
            "empty": error_empty_or_required,
        },
    )

    def validate_threads(self, data):
        limit = self.context["settings"].threads_per_page
        if len(data) > limit:
            message = ngettext(
                "No more than %(limit)s thread can be deleted at a single time.",
                "No more than %(limit)s threads can be deleted at a single time.",
                limit,
            )
            raise ValidationError(message % {"limit": limit})

        request = self.context["request"]
        viewmodel = self.context["viewmodel"]

        threads = []
        errors = []

        for thread_id in data:
            try:
                thread = viewmodel(request, thread_id).unwrap()
                allow_delete_thread(request.user_acl, thread)
                threads.append(thread)
            except PermissionDenied as e:
                errors.append(
                    {
                        "thread": {"id": thread.id, "title": thread.title},
                        "error": str(e),
                    }
                )
            except Http404 as e:
                pass  # skip invisible threads

        if errors:
            raise serializers.ValidationError({"details": errors})

        if len(threads) != len(data):
            raise ValidationError(
                _("One or more threads to delete could not be found.")
            )

        return threads


class MergeThreadSerializer(serializers.Serializer):
    other_thread = serializers.CharField(
        error_messages={"required": gettext_lazy("Enter link to new thread.")}
    )
    best_answer = serializers.IntegerField(
        required=False, error_messages={"invalid": gettext_lazy("Invalid choice.")}
    )
    poll = serializers.IntegerField(
        required=False, error_messages={"invalid": gettext_lazy("Invalid choice.")}
    )

    def validate_other_thread(self, data):
        request = self.context["request"]
        thread = self.context["thread"]
        viewmodel = self.context["viewmodel"]

        other_thread_id = get_thread_id_from_url(request, data)
        if not other_thread_id:
            raise ValidationError(_("This is not a valid thread link."))
        if other_thread_id == thread.pk:
            raise ValidationError(_("You can't merge thread with itself."))

        try:
            other_thread = viewmodel(request, other_thread_id).unwrap()
            allow_merge_thread(request.user_acl, other_thread, otherthread=True)
        except PermissionDenied as e:
            raise serializers.ValidationError(e)
        except Http404:
            raise ValidationError(
                _(
                    "The thread you have entered link to doesn't "
                    "exist or you don't have permission to see it."
                )
            )

        if not can_reply_thread(request.user_acl, other_thread):
            raise ValidationError(
                _("You can't merge this thread into thread you can't reply.")
            )

        return other_thread

    def validate(self, data):
        thread = self.context["thread"]
        other_thread = data["other_thread"]

        merge_conflict = MergeConflict(data, [thread, other_thread])
        merge_conflict.is_valid(raise_exception=True)
        data.update(merge_conflict.get_resolution())
        self.merge_conflict = merge_conflict.get_conflicting_fields()

        return data


class MergeThreadsSerializer(NewThreadSerializer):
    error_empty_or_required = gettext_lazy(
        "You have to select at least two threads to merge."
    )

    threads = serializers.ListField(
        allow_empty=False,
        min_length=2,
        child=serializers.IntegerField(
            error_messages={
                "invalid": gettext_lazy("One or more thread ids received were invalid.")
            }
        ),
        error_messages={
            "empty": error_empty_or_required,
            "null": error_empty_or_required,
            "required": error_empty_or_required,
            "min_length": error_empty_or_required,
        },
    )
    best_answer = serializers.IntegerField(
        required=False, error_messages={"invalid": gettext_lazy("Invalid choice.")}
    )
    poll = serializers.IntegerField(
        required=False, error_messages={"invalid": gettext_lazy("Invalid choice.")}
    )

    def validate_threads(self, data):
        limit = self.context["settings"].threads_per_page
        if len(data) > limit:
            message = ngettext(
                "No more than %(limit)s thread can be merged at a single time.",
                "No more than %(limit)s threads can be merged at a single time.",
                limit,
            )
            raise ValidationError(message % {"limit": limit})

        threads_tree_id = trees_map.get_tree_id_for_root(THREADS_ROOT_NAME)

        threads_queryset = (
            Thread.objects.filter(id__in=data, category__tree_id=threads_tree_id)
            .select_related("category")
            .order_by("-id")
        )

        user_acl = self.context["user_acl"]

        threads = []
        for thread in threads_queryset:
            add_acl_to_obj(user_acl, thread)
            if can_see_thread(user_acl, thread):
                threads.append(thread)

        if len(threads) != len(data):
            raise ValidationError(_("One or more threads to merge could not be found."))

        return threads



import os

from django.core.files import File
from django.urls import reverse

from ....admin.test import AdminTestCase
from ...datadownloads import request_user_data_download
from ...models import DataDownload
from ...test import create_test_user

APP_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
TESTFILES_DIR = os.path.join(APP_DIR, "tests", "testfiles")
TEST_FILE_PATH = os.path.join(TESTFILES_DIR, "avatar.png")


class DataDownloadAdminTests(AdminTestCase):
    def test_link_registered(self):
        """admin nav contains data downloads link"""
        response = self.client.get(reverse("misago:admin:users:index"))

        response = self.client.get(response["location"])
        self.assertContains(
            response, reverse("misago:admin:users:data-downloads:index")
        )

    def test_list_view(self):
        """data downloads list view returns 200"""
        response = self.client.get(reverse("misago:admin:users:data-downloads:index"))
        self.assertEqual(response.status_code, 302)

        view_url = response["location"]

        response = self.client.get(view_url)
        self.assertEqual(response.status_code, 200)

        request_user_data_download(self.user)
        response = self.client.get(view_url)
        self.assertEqual(response.status_code, 200)

    def test_expire_action(self):
        """expire action marks data download as expired and deletes its file"""
        data_download = request_user_data_download(self.user)

        with open(TEST_FILE_PATH, "rb") as upload:
            data_download.file = File(upload)
            data_download.save()

        self.assertIsNotNone(data_download.file)
        self.assertTrue(os.path.isfile(data_download.file.path))

        response = self.client.post(
            reverse("misago:admin:users:data-downloads:index"),
            data={"action": "expire", "selected_items": [data_download.pk]},
        )
        self.assertEqual(response.status_code, 302)

        updated_download = DataDownload.objects.get(pk=data_download.pk)
        self.assertEqual(updated_download.status, DataDownload.STATUS_EXPIRED)
        self.assertFalse(updated_download.file)

        self.assertFalse(os.path.isfile(data_download.file.path))

    def test_delete_action(self):
        """dele action deletes data download together with its file"""
        data_download = request_user_data_download(self.user)

        with open(TEST_FILE_PATH, "rb") as upload:
            data_download.file = File(upload)
            data_download.save()

        self.assertIsNotNone(data_download.file)
        self.assertTrue(os.path.isfile(data_download.file.path))

        response = self.client.post(
            reverse("misago:admin:users:data-downloads:index"),
            data={"action": "delete", "selected_items": [data_download.pk]},
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(DataDownload.objects.count(), 0)
        self.assertFalse(os.path.isfile(data_download.file.path))

    def test_request_view(self):
        """request data downloads view initializes new downloads"""
        response = self.client.get(reverse("misago:admin:users:data-downloads:request"))
        self.assertEqual(response.status_code, 200)

        other_user = create_test_user("OtherUser", "OtherUser@example.com")

        response = self.client.post(
            reverse("misago:admin:users:data-downloads:request"),
            data={
                "user_identifiers": "\n".join([self.user.username, other_user.email])
            },
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(DataDownload.objects.count(), 2)

    def test_request_view_empty_data(self):
        """request data downloads view handles empty data"""
        response = self.client.get(reverse("misago:admin:users:data-downloads:request"))
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            reverse("misago:admin:users:data-downloads:request"),
            data={"user_identifiers": ""},
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(DataDownload.objects.count(), 0)

    def test_request_view_user_not_found(self):
        """request data downloads view handles empty data"""
        response = self.client.get(reverse("misago:admin:users:data-downloads:request"))
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            reverse("misago:admin:users:data-downloads:request"),
            data={"user_identifiers": "not@found.com"},
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(DataDownload.objects.count(), 0)



from django.contrib.auth import get_user_model

from ...conf.test import override_dynamic_settings
from ..setupnewuser import set_default_subscription_options, setup_new_user

User = get_user_model()


def user(db):
    return User.objects.create_user("User", "user@example.com")


def test_default_avatar_is_set_for_user(dynamic_settings, user):
    setup_new_user(dynamic_settings, user)
    assert user.avatars
    assert user.avatar_set.exists()


def test_avatar_from_url_is_set_for_user(dynamic_settings, user):
    setup_new_user(dynamic_settings, user, avatar_url="https://placekitten.com/600/500")
    assert user.avatars
    assert user.avatar_set.exists()


def test_default_started_threads_subscription_option_is_set_for_user(
    dynamic_settings, user
):
    with override_dynamic_settings(subscribe_start="no"):
        set_default_subscription_options(dynamic_settings, user)
        assert user.subscribe_to_started_threads == User.SUBSCRIPTION_NONE

    with override_dynamic_settings(subscribe_start="watch"):
        set_default_subscription_options(dynamic_settings, user)
        assert user.subscribe_to_started_threads == User.SUBSCRIPTION_NOTIFY

    with override_dynamic_settings(subscribe_start="watch_email"):
        set_default_subscription_options(dynamic_settings, user)
        assert user.subscribe_to_started_threads == User.SUBSCRIPTION_ALL


def test_default_replied_threads_subscription_option_is_set_for_user(
    dynamic_settings, user
):
    with override_dynamic_settings(subscribe_reply="no"):
        set_default_subscription_options(dynamic_settings, user)
        assert user.subscribe_to_replied_threads == User.SUBSCRIPTION_NONE

    with override_dynamic_settings(subscribe_reply="watch"):
        set_default_subscription_options(dynamic_settings, user)
        assert user.subscribe_to_replied_threads == User.SUBSCRIPTION_NOTIFY

    with override_dynamic_settings(subscribe_reply="watch_email"):
        set_default_subscription_options(dynamic_settings, user)
        assert user.subscribe_to_replied_threads == User.SUBSCRIPTION_ALL


def test_if_user_ip_is_available_audit_trail_is_created_for_user(dynamic_settings):
    user = User.objects.create_user(
        "User", "user@example.com", joined_from_ip="0.0.0.0"
    )
    setup_new_user(dynamic_settings, user)
    assert user.audittrail_set.count() == 1


def test_if_user_ip_is_not_available_audit_trail_is_not_created(dynamic_settings, user):
    setup_new_user(dynamic_settings, user)
    assert user.audittrail_set.exists() is False



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots[
    "test_simple_url_to_file_is_replaced_with_valid_url 1"
] = '.page-header { background-image: url("/media/themes/themedir/media/test.357c2ee3.png"); }'

snapshots[
    "test_relative_url_to_file_is_replaced_with_valid_url 1"
] = '.page-header { background-image: url("/media/themes/themedir/media/test.357c2ee3.png"); }'

snapshots[
    "test_url_to_file_from_create_react_app_is_replaced_with_valid_url 1"
] = '.page-header { background-image: url("/media/themes/themedir/media/test.357c2ee3.png"); }'

snapshots[
    "test_quoted_url_to_file_is_replaced_with_valid_url 1"
] = '.page-header { background-image: url("/media/themes/themedir/media/test.357c2ee3.png"); }'

snapshots[
    "test_single_quoted_url_to_file_is_replaced_with_valid_url 1"
] = '.page-header { background-image: url("/media/themes/themedir/media/test.357c2ee3.png"); }'

snapshots[
    "test_css_file_with_multiple_different_urls_is_correctly_replaced 1"
] = """.page-header { background-image: url(http://cdn.example.com/bg.png); }
.container { background-image: url("/media/themes/themedir/media/test.357c2ee3.png"); }
.alert { background-image: url("g"); }"""



import pytest

from ..models import SocialAuthProvider


@pytest.fixture
def provider(db):
    return SocialAuthProvider.objects.create(
        provider="facebook",
        is_active=True,
        order=0,
        settings={"key": "test-key", "secret": "test-secret"},
    )


@pytest.fixture
def disabled_provider(db):
    return SocialAuthProvider.objects.create(
        provider="google", is_active=False, order=0
    )



from functools import wraps

from django.contrib import auth
from django.contrib.auth.password_validation import validate_password
from django.core.exceptions import PermissionDenied, ValidationError
from django.utils.translation import gettext as _
from django.views.decorators.csrf import csrf_protect
from rest_framework import status
from rest_framework.decorators import api_view, permission_classes
from rest_framework.response import Response

from ...conf import settings
from ...core.decorators import require_dict_data
from ...core.mail import mail_user
from ..bans import get_user_ban
from ..forms.auth import AuthenticationForm, ResendActivationForm, ResetPasswordForm
from ..serializers import AnonymousUserSerializer, AuthenticatedUserSerializer
from ..tokens import (
    is_password_change_token_valid,
    make_activation_token,
    make_password_change_token,
)
from .rest_permissions import UnbannedAnonOnly, UnbannedOnly

User = auth.get_user_model()


def gateway(request):
    if request.method == "POST":
        return login(request)
    return session_user(request)


def check_delegated_auth(f):
    @wraps(f)
    def view_disabled_if_auth_is_delegated(request, *args, **kwargs):
        if request.settings.enable_oauth2_client:
            raise PermissionDenied(
                _(
                    "This feature has been disabled. "
                    "Please use %(provider)s to sign in."
                )
                % {"provider": request.settings.oauth2_provider}
            )

        return f(request, *args, **kwargs)

    return view_disabled_if_auth_is_delegated


@api_view(["POST"])
@permission_classes((UnbannedAnonOnly,))
@check_delegated_auth
@csrf_protect
@require_dict_data
def login(request):
    """
    POST /auth/ with CSRF, username and password
    will attempt to authenticate new user
    """
    form = AuthenticationForm(request, data=request.data)
    if form.is_valid():
        auth.login(request, form.user_cache)
        return Response(AuthenticatedUserSerializer(form.user_cache).data)
    return Response(form.get_errors_dict(), status=status.HTTP_400_BAD_REQUEST)


@api_view()
def session_user(request):
    """GET /auth/ will return current auth user, either User or AnonymousUser"""
    if request.user.is_authenticated:
        serializer = AuthenticatedUserSerializer
    else:
        serializer = AnonymousUserSerializer

    serialized_user = serializer(request.user, context={"acl": request.user_acl}).data
    return Response(serialized_user)


@api_view(["GET"])
def get_criteria(request):
    """GET /auth/criteria/ will return password and username criteria for accounts"""
    criteria = {
        "username": {
            "min_length": request.settings.username_length_min,
            "max_length": request.settings.username_length_max,
        },
        "password": [],
    }

    for validator in settings.AUTH_PASSWORD_VALIDATORS:
        validator_dict = {"name": validator["NAME"].split(".")[-1]}
        validator_dict.update(validator.get("OPTIONS", {}))
        criteria["password"].append(validator_dict)

    return Response(criteria)


@api_view(["POST"])
@permission_classes((UnbannedAnonOnly,))
@check_delegated_auth
@csrf_protect
@require_dict_data
def send_activation(request):
    """
    POST /auth/send-activation/ with CSRF token and email
    will mail account activation link to requester
    """
    form = ResendActivationForm(request.data)
    if not form.is_valid():
        return Response(form.get_errors_dict(), status=status.HTTP_400_BAD_REQUEST)

    requesting_user = form.user_cache

    mail_subject = _("Activate %(user)s account on %(forum_name)s forums") % {
        "user": requesting_user.username,
        "forum_name": request.settings.forum_name,
    }

    mail_user(
        requesting_user,
        mail_subject,
        "misago/emails/activation/by_user",
        context={
            "activation_token": make_activation_token(requesting_user),
            "settings": request.settings,
        },
    )

    return Response(
        {"username": form.user_cache.username, "email": form.user_cache.email}
    )


@api_view(["POST"])
@permission_classes((UnbannedOnly,))
@check_delegated_auth
@csrf_protect
@require_dict_data
def send_password_form(request):
    """
    POST /auth/send-password-form/ with CSRF token and email
    will mail change password form link to requester
    """
    form = ResetPasswordForm(request.data)
    if not form.is_valid():
        return Response(form.get_errors_dict(), status=status.HTTP_400_BAD_REQUEST)

    requesting_user = form.user_cache

    mail_subject = _("Change %(user)s password on %(forum_name)s forums") % {
        "user": requesting_user.username,
        "forum_name": request.settings.forum_name,
    }

    confirmation_token = make_password_change_token(requesting_user)

    mail_user(
        requesting_user,
        mail_subject,
        "misago/emails/change_password_form_link",
        context={
            "confirmation_token": confirmation_token,
            "settings": request.settings,
        },
    )

    return Response(
        {"username": form.user_cache.username, "email": form.user_cache.email}
    )


class PasswordChangeFailed(Exception):
    pass


@api_view(["POST"])
@permission_classes((UnbannedOnly,))
@check_delegated_auth
@csrf_protect
@require_dict_data
def change_forgotten_password(request, pk, token):
    """
    POST /auth/change-password/user/token/ with CSRF and new password
    will change forgotten password
    """
    invalid_message = _("Form link is invalid. Please try again.")
    expired_message = _("Your link has expired. Please request new one.")

    try:
        try:
            user = User.objects.get(pk=pk, is_active=True)
        except User.DoesNotExist:
            raise PasswordChangeFailed(invalid_message)

        if request.user.is_authenticated and request.user.id != user.id:
            raise PasswordChangeFailed(invalid_message)
        if not is_password_change_token_valid(user, token):
            raise PasswordChangeFailed(invalid_message)

        if user.requires_activation:
            raise PasswordChangeFailed(expired_message)
        if get_user_ban(user, request.cache_versions):
            raise PasswordChangeFailed(expired_message)
    except PasswordChangeFailed as e:
        return Response({"detail": e.args[0]}, status=status.HTTP_400_BAD_REQUEST)

    try:
        new_password = request.data.get("password", "")
        validate_password(new_password, user=user)
        user.set_password(new_password)
        user.save()
    except ValidationError as e:
        return Response({"detail": e.messages[0]}, status=status.HTTP_400_BAD_REQUEST)

    return Response({"username": user.username})



import os

from django.utils.module_loading import import_string
from PIL import Image, ImageColor, ImageDraw, ImageFont

from . import store
from ...conf import settings

COLOR_WHEEL = (
    "#d32f2f",
    "#c2185b",
    "#7b1fa2",
    "#512da8",
    "#303f9f",
    "#1976d2",
    "#0288D1",
    "#0288d1",
    "#0097a7",
    "#00796b",
    "#388e3c",
    "#689f38",
    "#afb42b",
    "#fbc02d",
    "#ffa000",
    "#f57c00",
    "#e64a19",
)
COLOR_WHEEL_LEN = len(COLOR_WHEEL)
FONT_FILE = os.path.join(os.path.dirname(__file__), "font.ttf")


def set_avatar(user):
    drawer_function = import_string(settings.MISAGO_DYNAMIC_AVATAR_DRAWER)

    image = drawer_function(user)
    store.store_new_avatar(user, image)


def draw_default(user):
    """default avatar drawer that draws username's first letter on color"""
    image_size = max(settings.MISAGO_AVATARS_SIZES)

    image = Image.new("RGBA", (image_size, image_size), 0)
    image = draw_avatar_bg(user, image)
    image = draw_avatar_flavour(user, image)

    return image


def draw_avatar_bg(user, image):
    image_size = image.size

    color_index = user.pk - COLOR_WHEEL_LEN * (user.pk // COLOR_WHEEL_LEN)
    main_color = COLOR_WHEEL[color_index]

    rgb = ImageColor.getrgb(main_color)

    bg_drawer = ImageDraw.Draw(image)
    bg_drawer.rectangle([(0, 0), image_size], rgb)

    return image


def draw_avatar_flavour(user, image):
    string = user.username[0]

    image_size = image.size[0]

    size = int(image_size * 0.7)
    font = ImageFont.truetype(FONT_FILE, size=size)

    text_box = font.getbbox(string)
    text_pos = ((image_size - text_box[2]) / 2, (image_size - text_box[3]) / 2)

    writer = ImageDraw.Draw(image)
    writer.text(text_pos, string, font=font)

    return image



from django.urls import reverse

from ...conf.test import override_dynamic_settings
from ...test import assert_contains


@override_dynamic_settings(google_site_verification=None)
def test_verification_page_returns_404_if_verification_token_is_not_set(db, client):
    verification_link = reverse(
        "misago:google-site-verification", kwargs={"token": "t0k3n"}
    )
    response = client.get(verification_link)
    assert response.status_code == 404


@override_dynamic_settings(google_site_verification="t0k3n")
def test_verification_page_returns_404_if_verification_token_is_invalid(db, client):
    verification_link = reverse(
        "misago:google-site-verification", kwargs={"token": "inv4l1d"}
    )
    response = client.get(verification_link)
    assert response.status_code == 404


@override_dynamic_settings(google_site_verification="t0k3n")
def test_verification_page_returns_200_if_verification_token_is_correct(db, client):
    verification_link = reverse(
        "misago:google-site-verification", kwargs={"token": "t0k3n"}
    )
    response = client.get(verification_link)
    assert response.status_code == 200


@override_dynamic_settings(google_site_verification="t0k3n")
def test_verification_page_contains_token(db, client):
    verification_link = reverse(
        "misago:google-site-verification", kwargs={"token": "t0k3n"}
    )
    response = client.get(verification_link)
    assert_contains(response, "googlet0k3n.html")



import pytest

from ..parser import parse


@pytest.mark.parametrize(
    "text",
    [
        pytest.param("!(http://somewhere.com/image.jpg)", id="base"),
        pytest.param("! (space)", id="space-one-word"),
        pytest.param("! (space with other words)", id="space-multiple-words"),
        pytest.param(
            "Text before exclamation mark!(http://somewhere.com/image.jpg)",
            id="text-before-mark",
        ),
        pytest.param(
            "Text before with space in between! (sometext)", id="text-before-with-space"
        ),
    ],
)
def test_short_image_markdown(request_mock, user, snapshot, text):
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])



from .cache import get_socialauth_cache, set_socialauth_cache
from .models import SocialAuthProvider
from .providers import providers


def get_enabled_providers(cache_versions):
    data = get_socialauth_cache(cache_versions)
    if data is None:
        data = get_providers_from_db()
        set_socialauth_cache(cache_versions, data)
    for provider, options in data.items():
        options["auth_backend"] = providers.get_auth_backend(provider)
    return data


def get_providers_from_db():
    data = {}
    for provider in SocialAuthProvider.objects.filter(is_active=True):
        data[provider.pk] = {
            "pk": provider.pk,
            "name": providers.get_name(provider.pk),
            "settings": get_provider_settings(provider),
            "auth_backend": None,
            "button_text": provider.button_text,
            "button_color": provider.button_color,
        }
    return data


def get_provider_settings(provider):
    settings = {}
    if providers.get_settings(provider.pk):
        for key, value in providers.get_settings(provider.pk).items():
            settings[key.upper()] = value
    if provider.settings:
        for key, value in provider.settings.items():
            settings[key.upper()] = value
    return settings



from io import BytesIO


from PIL import Image
from django import forms
from django.core.files.base import ContentFile
from django.utils.translation import gettext_lazy as _

from ...core.utils import get_file_hash
from ...core.validators import validate_image_square
from ..models import Icon

FAVICON_MIN_SIZE = 48
FAVICON_SIZES = ((16, 16), (32, 32), (48, 48))
APPLE_TOUCH_MIN_SIZE = 180
VALID_MIME = ("image/gif", "image/jpeg", "image/png")


class IconsForm(forms.Form):
    favicon = forms.ImageField(
        label=_("Upload image"),
        help_text=_("Uploaded image should be a square that is 48px wide and tall."),
        required=False,
    )
    favicon_delete = forms.BooleanField(label=_("Delete custom icon"), required=False)

    apple_touch_icon = forms.ImageField(
        label=_("Upload image"),
        help_text=_("Uploaded image should be square at least 180px wide and tall."),
        required=False,
    )
    apple_touch_icon_delete = forms.BooleanField(
        label=_("Delete custom icon"), required=False
    )

    def clean_favicon(self):
        upload = self.cleaned_data.get("favicon")
        if not upload or upload == self.initial.get("favicon"):
            return None

        validate_image_square(upload.image)
        validate_image_dimensions(upload.image, FAVICON_MIN_SIZE)
        validate_image_mime_type(upload)

        return upload

    def clean_apple_touch_icon(self):
        upload = self.cleaned_data.get("apple_touch_icon")
        if not upload or upload == self.initial.get("apple_touch_icon"):
            return None

        validate_image_square(upload.image)
        validate_image_dimensions(upload.image, APPLE_TOUCH_MIN_SIZE)
        validate_image_mime_type(upload)

        return upload

    def save(self):
        if self.cleaned_data.get("favicon"):
            self.save_favicon(self.cleaned_data["favicon"])
        elif self.cleaned_data.get("favicon_delete"):
            self.delete_icons(Icon.FAVICON_TYPES)

        if self.cleaned_data.get("apple_touch_icon"):
            self.save_apple_touch_icon(self.cleaned_data["apple_touch_icon"])
        elif self.cleaned_data.get("apple_touch_icon_delete"):
            self.delete_icons([Icon.TYPE_APPLE_TOUCH_ICON])

    def save_favicon(self, image):
        self.delete_icons(Icon.FAVICON_TYPES)
        save_favicon(image)
        save_icon(image, (32, 32), Icon.TYPE_FAVICON_32)
        save_icon(image, (16, 16), Icon.TYPE_FAVICON_16)

    def save_apple_touch_icon(self, image):
        self.delete_icons([Icon.TYPE_APPLE_TOUCH_ICON])
        save_icon(image, (180, 180), Icon.TYPE_APPLE_TOUCH_ICON)

    def delete_icons(self, icon_types):
        for icon in Icon.objects.filter(type__in=icon_types):
            icon.delete()


def save_favicon(image):
    icon = Image.open(image)

    buffer = BytesIO()
    icon.save(buffer, "ICO", sizes=FAVICON_SIZES)
    buffer.seek(0)

    icon_file = ContentFile(buffer.read())
    icon_file.name = "%s.%s.ico" % ("favicon", get_file_hash(icon_file))

    Icon.objects.create(type=Icon.TYPE_FAVICON, image=icon_file, size=icon_file.size)


def save_icon(image, size, icon_type):
    icon = Image.open(image)
    icon.thumbnail(size)

    buffer = BytesIO()
    icon.save(buffer, "PNG")
    buffer.seek(0)

    icon_file = ContentFile(buffer.read())
    icon_file.name = "%s.%s.png" % (
        icon_type.replace("_", "-"),
        get_file_hash(icon_file),
    )

    Icon.objects.create(type=icon_type, image=icon_file, size=icon_file.size)


def validate_image_dimensions(image, size):
    if image.width < size:
        raise forms.ValidationError(
            _("Uploaded image's edge should be at least %(size)s pixels long.")
            % {"size": size}
        )


def validate_image_mime_type(upload):
    if upload.content_type not in VALID_MIME:
        raise forms.ValidationError(_("Uploaded image was not gif, jpeg or png."))



from django.urls import path
from django.utils.translation import gettext_lazy as _

from .views import (
    ActivateTheme,
    DeleteTheme,
    DeleteThemeCss,
    DeleteThemeMedia,
    EditTheme,
    EditThemeCss,
    EditThemeCssLink,
    MoveThemeCssDown,
    MoveThemeCssUp,
    NewTheme,
    NewThemeCss,
    NewThemeCssLink,
    ExportTheme,
    ImportTheme,
    ThemeAssets,
    ThemesList,
    UploadThemeCss,
    UploadThemeMedia,
)


class MisagoAdminExtension:
    def register_urlpatterns(self, urlpatterns):
        # Themes
        urlpatterns.namespace("themes/", "themes")
        urlpatterns.patterns(
            "themes",
            path("", ThemesList.as_view(), name="index"),
            path("new/", NewTheme.as_view(), name="new"),
            path("edit/<int:pk>/", EditTheme.as_view(), name="edit"),
            path("delete/<int:pk>/", DeleteTheme.as_view(), name="delete"),
            path("activate/<int:pk>/", ActivateTheme.as_view(), name="activate"),
            path("export/<int:pk>/", ExportTheme.as_view(), name="export"),
            path("import/", ImportTheme.as_view(), name="import"),
            path("assets/<int:pk>/", ThemeAssets.as_view(), name="assets"),
            path(
                "assets/<int:pk>/delete-css/",
                DeleteThemeCss.as_view(),
                name="delete-css",
            ),
            path(
                "assets/<int:pk>/delete-media/",
                DeleteThemeMedia.as_view(),
                name="delete-media",
            ),
            path(
                "assets/<int:pk>/upload-css/",
                UploadThemeCss.as_view(),
                name="upload-css",
            ),
            path(
                "assets/<int:pk>/upload-media/",
                UploadThemeMedia.as_view(),
                name="upload-media",
            ),
            path(
                "assets/<int:pk>/move-css-down/<int:css_pk>/",
                MoveThemeCssDown.as_view(),
                name="move-css-down",
            ),
            path(
                "assets/<int:pk>/move-css-up/<int:css_pk>/",
                MoveThemeCssUp.as_view(),
                name="move-css-up",
            ),
            path(
                "assets/<int:pk>/new-css/",
                NewThemeCss.as_view(),
                name="new-css-file",
            ),
            path(
                "assets/<int:pk>/edit-css/<int:css_pk>/",
                EditThemeCss.as_view(),
                name="edit-css-file",
            ),
            path(
                "assets/<int:pk>/new-css-link/",
                NewThemeCssLink.as_view(),
                name="new-css-link",
            ),
            path(
                "assets/<int:pk>/edit-css-link/<int:css_pk>/",
                EditThemeCssLink.as_view(),
                name="edit-css-link",
            ),
        )

    def register_navigation_nodes(self, site):
        site.add_node(
            name=_("Themes"),
            icon="fa fa-paint-brush",
            after="attachments:index",
            namespace="themes",
        )



from django.urls import reverse

from .models import Agreement
from .utils import get_parsed_agreement_text, get_required_user_agreement

# fixme: rename this context processor to more suitable name
def legal_links(request):
    agreements = Agreement.objects.get_agreements()

    legal_context = {
        "TERMS_OF_SERVICE_ID": None,
        "TERMS_OF_SERVICE_URL": None,
        "PRIVACY_POLICY_ID": None,
        "PRIVACY_POLICY_URL": None,
        "misago_agreement": None,
    }

    terms_of_service = agreements.get(Agreement.TYPE_TOS)
    if terms_of_service:
        legal_context["TERMS_OF_SERVICE_ID"] = terms_of_service["id"]
        if terms_of_service["link"]:
            legal_context["TERMS_OF_SERVICE_URL"] = terms_of_service["link"]
        elif terms_of_service["text"]:
            legal_context["TERMS_OF_SERVICE_URL"] = reverse("misago:terms-of-service")

    privacy_policy = agreements.get(Agreement.TYPE_PRIVACY)
    if privacy_policy:
        legal_context["PRIVACY_POLICY_ID"] = privacy_policy["id"]
        if privacy_policy["link"]:
            legal_context["PRIVACY_POLICY_URL"] = privacy_policy["link"]
        elif privacy_policy["text"]:
            legal_context["PRIVACY_POLICY_URL"] = reverse("misago:privacy-policy")

    if legal_context:
        request.frontend_context.update(legal_context)

    required_agreement = get_required_user_agreement(request.user, agreements)
    if required_agreement:
        request.frontend_context["REQUIRED_AGREEMENT_API"] = reverse(
            "misago:api:submit-agreement", kwargs={"pk": required_agreement.pk}
        )

        legal_context["misago_agreement"] = {
            "type": required_agreement.get_type_display(),
            "title": required_agreement.get_final_title(),
            "link": required_agreement.link,
            "text": get_parsed_agreement_text(request, required_agreement),
        }

    return legal_context



from ..parser import parse


def test_single_line_code(request_mock, user, snapshot):
    text = '[code]echo("Hello!");[/code]'
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_multi_line_code(request_mock, user, snapshot):
    text = """
[code]
<script>
alert("!")
</script>
[/code]
    """
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_code_with_language_parameter(request_mock, user, snapshot):
    text = '[code=php]echo("Hello!");[/code]'
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_code_with_quoted_language_parameter(request_mock, user, snapshot):
    text = '[code="php"]echo("Hello!");[/code]'
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_code_block_disables_parsing(request_mock, user, snapshot):
    text = "[code]Dolor [b]met.[/b][/code]"
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])



import os
from collections import OrderedDict

from django.core.files import File
from django.test import TestCase
from django.utils import timezone

from ...conf import settings
from ..datadownloads.dataarchive import (
    FILENAME_MAX_LEN,
    DataArchive,
    trim_long_filename,
)
from ..test import AuthenticatedUserTestCase

DATA_DOWNLOADS_WORKING_DIR = settings.MISAGO_USER_DATA_DOWNLOADS_WORKING_DIR
TESTFILES_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "testfiles")
TEST_AVATAR_PATH = os.path.join(TESTFILES_DIR, "avatar.png")


class DataArchiveTests(AuthenticatedUserTestCase):
    def test_enter_without_dirs(self):
        """data archive doesn't touch filesystem on init"""
        archive = DataArchive(self.user, DATA_DOWNLOADS_WORKING_DIR)

        self.assertEqual(archive.user, self.user)
        self.assertEqual(archive.working_dir_path, DATA_DOWNLOADS_WORKING_DIR)

        self.assertIsNone(archive.tmp_dir_path)
        self.assertIsNone(archive.data_dir_path)

    def test_context_life_cycle(self):
        """object creates valid tmp directory on enter and cleans on exit"""
        tmp_dir_path = None
        data_dir_path = None

        with DataArchive(self.user, DATA_DOWNLOADS_WORKING_DIR) as archive:
            self.assertTrue(os.path.exists(archive.tmp_dir_path))
            self.assertTrue(os.path.exists(archive.data_dir_path))

            working_dir = str(DATA_DOWNLOADS_WORKING_DIR)
            tmp_dir_path = str(archive.tmp_dir_path)
            data_dir_path = str(archive.data_dir_path)

            self.assertTrue(tmp_dir_path.startswith(working_dir))
            self.assertTrue(data_dir_path.startswith(working_dir))

            self.assertTrue(data_dir_path.startswith(working_dir))
            self.assertTrue(data_dir_path.startswith(tmp_dir_path))

        self.assertIsNone(archive.tmp_dir_path)
        self.assertIsNone(archive.data_dir_path)

        self.assertFalse(os.path.exists(tmp_dir_path))
        self.assertFalse(os.path.exists(data_dir_path))

    def test_add_text_str(self):
        """add_dict method creates text file with string"""
        with DataArchive(self.user, DATA_DOWNLOADS_WORKING_DIR) as archive:
            data_to_write = "Hello, łorld!"
            file_path = archive.add_text("testfile", data_to_write)
            self.assertTrue(os.path.isfile(file_path))

            valid_output_path = os.path.join(archive.data_dir_path, "testfile.txt")
            self.assertEqual(file_path, valid_output_path)

            with open(file_path, "r") as fp:
                saved_data = fp.read().strip()
                self.assertEqual(saved_data, data_to_write)

    def test_add_text_int(self):
        """add_dict method creates text file with int"""
        with DataArchive(self.user, DATA_DOWNLOADS_WORKING_DIR) as archive:
            data_to_write = 1234
            file_path = archive.add_text("testfile", data_to_write)
            self.assertTrue(os.path.isfile(file_path))

            valid_output_path = os.path.join(archive.data_dir_path, "testfile.txt")
            self.assertEqual(file_path, valid_output_path)

            with open(file_path, "r") as fp:
                saved_data = fp.read().strip()
                self.assertEqual(saved_data, str(data_to_write))

    def test_add_dict(self):
        """add_dict method creates text file from dict"""
        with DataArchive(self.user, DATA_DOWNLOADS_WORKING_DIR) as archive:
            data_to_write = {"first": "łorld!", "second": "łup!"}
            file_path = archive.add_dict("testfile", data_to_write)
            self.assertTrue(os.path.isfile(file_path))

            valid_output_path = os.path.join(archive.data_dir_path, "testfile.txt")
            self.assertEqual(file_path, valid_output_path)

            with open(file_path, "r") as fp:
                saved_data = fp.read().strip()
                # order of dict items in py<3.6 is non-deterministic
                # making testing for exact match a mistake
                self.assertIn("first: łorld!", saved_data)
                self.assertIn("second: łup!", saved_data)

    def test_add_dict_ordered(self):
        """add_dict method creates text file form ordered dict"""
        with DataArchive(self.user, DATA_DOWNLOADS_WORKING_DIR) as archive:
            data_to_write = OrderedDict((("first", "łorld!"), ("second", "łup!")))
            file_path = archive.add_dict("testfile", data_to_write)
            self.assertTrue(os.path.isfile(file_path))

            valid_output_path = os.path.join(archive.data_dir_path, "testfile.txt")
            self.assertEqual(file_path, valid_output_path)

            with open(file_path, "r") as fp:
                saved_data = fp.read().strip()
                self.assertEqual(saved_data, "first: łorld!\nsecond: łup!")

    def test_add_model_file(self):
        """add_model_file method adds model file"""
        with open(TEST_AVATAR_PATH, "rb") as avatar:
            self.user.avatar_tmp = File(avatar)
            self.user.save()

        with DataArchive(self.user, DATA_DOWNLOADS_WORKING_DIR) as archive:
            file_path = archive.add_model_file(self.user.avatar_tmp)

            self.assertTrue(os.path.isfile(file_path))

            data_dir_path = str(archive.data_dir_path)
            self.assertTrue(str(file_path).startswith(data_dir_path))

    def test_add_model_file_empty(self):
        """add_model_file method is noop if model field is empty"""
        with DataArchive(self.user, DATA_DOWNLOADS_WORKING_DIR) as archive:
            file_path = archive.add_model_file(self.user.avatar_tmp)

            self.assertIsNone(file_path)
            self.assertFalse(os.listdir(archive.data_dir_path))

    def test_add_model_file_prefixed(self):
        """add_model_file method adds model file with prefix"""
        with open(TEST_AVATAR_PATH, "rb") as avatar:
            self.user.avatar_tmp = File(avatar)
            self.user.save()

        with DataArchive(self.user, DATA_DOWNLOADS_WORKING_DIR) as archive:
            file_path = archive.add_model_file(self.user.avatar_tmp, prefix="prefix")

            self.assertTrue(os.path.isfile(file_path))

            data_dir_path = str(archive.data_dir_path)
            self.assertTrue(str(file_path).startswith(data_dir_path))

            filename = os.path.basename(self.user.avatar_tmp.name)
            target_filename = "prefix-%s" % filename
            self.assertTrue(str(file_path).endswith(target_filename))

    def test_make_final_path_no_kwargs(self):
        """make_final_path returns data_dir_path if no kwargs are set"""
        with DataArchive(self.user, DATA_DOWNLOADS_WORKING_DIR) as archive:
            final_path = archive.make_final_path()
            self.assertEqual(final_path, archive.data_dir_path)

    def test_make_final_path_directory(self):
        """make_final_path returns path including directory name"""
        with DataArchive(self.user, DATA_DOWNLOADS_WORKING_DIR) as archive:
            final_path = archive.make_final_path(directory="test-directory")
            valid_path = os.path.join(archive.data_dir_path, "test-directory")
            self.assertEqual(final_path, valid_path)

    def test_make_final_path_date(self):
        """make_final_path returns path including date segments"""
        with DataArchive(self.user, DATA_DOWNLOADS_WORKING_DIR) as archive:
            now = timezone.now().date()
            final_path = archive.make_final_path(date=now)

            valid_path = os.path.join(
                archive.data_dir_path,
                now.strftime("%Y"),
                now.strftime("%m"),
                now.strftime("%d"),
            )

            self.assertEqual(final_path, valid_path)

    def test_make_final_path_datetime(self):
        """make_final_path returns path including date segments"""
        with DataArchive(self.user, DATA_DOWNLOADS_WORKING_DIR) as archive:
            now = timezone.now()
            final_path = archive.make_final_path(date=now)

            valid_path = os.path.join(
                archive.data_dir_path,
                now.strftime("%Y"),
                now.strftime("%m"),
                now.strftime("%d"),
            )

            self.assertEqual(final_path, valid_path)

    def test_make_final_path_both_kwargs(self):
        """make_final_path raises value error if both date and directory are set"""
        with DataArchive(self.user, DATA_DOWNLOADS_WORKING_DIR) as archive:
            expected_message = "date and directory arguments are mutually exclusive"
            with self.assertRaisesMessage(ValueError, expected_message):
                archive.make_final_path(date=timezone.now(), directory="test")

    def test_get_file(self):
        """get_file returns django file"""
        django_file = None

        with open(TEST_AVATAR_PATH, "rb") as avatar:
            self.user.avatar_tmp = File(avatar)
            self.user.save()

        with DataArchive(self.user, DATA_DOWNLOADS_WORKING_DIR) as archive:
            archive.add_model_file(self.user.avatar_tmp)

            django_file = archive.get_file()
            archive_path = archive.file_path

            self.assertIsNotNone(archive_path)
            self.assertEqual(django_file.name, archive.file_path)
            self.assertFalse(django_file.closed)

        self.assertIsNone(archive.file)
        self.assertIsNone(archive.file_path)
        self.assertTrue(django_file.closed)


class TrimLongFilenameTests(TestCase):
    def test_trim_short_filename(self):
        """trim_too_long_filename returns short filename as it is"""
        filename = "filename.jpg"
        trimmed_filename = trim_long_filename(filename)
        self.assertEqual(trimmed_filename, filename)

    def test_trim_too_long_filename(self):
        """trim_too_long_filename trims filename if its longer than allowed"""
        filename = "filename"
        extension = ".jpg"
        long_filename = "%s%s" % (filename * 10, extension)

        trimmed_filename = trim_long_filename(long_filename)

        self.assertEqual(len(trimmed_filename), FILENAME_MAX_LEN)
        self.assertTrue(trimmed_filename.startswith(filename))
        self.assertTrue(trimmed_filename.endswith(extension))



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots[
    "test_short_image_markdown[base] 1"
] = '<p><img alt="somewhere.com/image.jpg" src="http://somewhere.com/image.jpg" /></p>'

snapshots[
    "test_short_image_markdown[space-multiple-words] 1"
] = "<p>! (space with other words)</p>"

snapshots["test_short_image_markdown[space-one-word] 1"] = "<p>! (space)</p>"

snapshots[
    "test_short_image_markdown[text-before-mark] 1"
] = '<p>Text before exclamation mark<img alt="somewhere.com/image.jpg" src="http://somewhere.com/image.jpg" /></p>'

snapshots[
    "test_short_image_markdown[text-before-with-space] 1"
] = "<p>Text before with space in between! (sometext)</p>"



import json

from django.urls import reverse

from .. import test
from ...categories.models import Category
from ...conf.test import override_dynamic_settings
from ...readtracker import poststracker
from ...users.test import AuthenticatedUserTestCase
from ..models import Thread
from ..test import patch_category_acl, patch_other_category_acl


class ThreadPostMoveApiTestCase(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")
        self.thread = test.post_thread(category=self.category)

        self.api_link = reverse(
            "misago:api:thread-post-move", kwargs={"thread_pk": self.thread.pk}
        )

        Category(name="Other category", slug="other-category").insert_at(
            self.category, position="last-child", save=True
        )
        self.other_category = Category.objects.get(slug="other-category")

    def test_anonymous_user(self):
        """you need to authenticate to move posts"""
        self.logout_user()

        response = self.client.post(
            self.api_link, json.dumps({}), content_type="application/json"
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This action is not available to guests."}
        )

    @patch_category_acl({"can_move_posts": True})
    def test_invalid_data(self):
        """api handles post that is invalid type"""
        response = self.client.post(
            self.api_link, "[]", content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "Invalid data. Expected a dictionary, but got list."},
        )

        response = self.client.post(
            self.api_link, "123", content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "Invalid data. Expected a dictionary, but got int."},
        )

        response = self.client.post(
            self.api_link, '"string"', content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "Invalid data. Expected a dictionary, but got str."},
        )

        response = self.client.post(
            self.api_link, "malformed", content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "JSON parse error - Expecting value: line 1 column 1 (char 0)"},
        )

    @patch_category_acl({"can_move_posts": False})
    def test_no_permission(self):
        """api validates permission to move"""
        response = self.client.post(
            self.api_link, json.dumps({}), content_type="application/json"
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't move posts in this thread."}
        )

    @patch_category_acl({"can_move_posts": True})
    def test_move_no_new_thread_url(self):
        """api validates if new thread url was given"""
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Enter link to new thread."})

    @patch_category_acl({"can_move_posts": True})
    def test_invalid_new_thread_url(self):
        """api validates new thread url"""
        response = self.client.post(
            self.api_link, {"new_thread": self.user.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "This is not a valid thread link."}
        )

    @patch_category_acl({"can_move_posts": True})
    def test_current_new_thread_url(self):
        """api validates if new thread url points to current thread"""
        response = self.client.post(
            self.api_link, {"new_thread": self.thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "Thread to move posts to is same as current one."},
        )

    @patch_other_category_acl({"can_see": False})
    @patch_category_acl({"can_move_posts": True})
    def test_other_thread_exists(self):
        """api validates if other thread exists"""
        other_thread = test.post_thread(self.other_category)

        response = self.client.post(
            self.api_link, {"new_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "detail": (
                    "The thread you have entered link to doesn't exist "
                    "or you don't have permission to see it."
                )
            },
        )

    @patch_other_category_acl({"can_browse": False})
    @patch_category_acl({"can_move_posts": True})
    def test_other_thread_is_invisible(self):
        """api validates if other thread is visible"""
        other_thread = test.post_thread(self.other_category)

        response = self.client.post(
            self.api_link, {"new_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "detail": (
                    "The thread you have entered link to doesn't exist "
                    "or you don't have permission to see it."
                )
            },
        )

    @patch_other_category_acl({"can_reply_threads": False})
    @patch_category_acl({"can_move_posts": True})
    def test_other_thread_isnt_replyable(self):
        """api validates if other thread can be replied"""
        other_thread = test.post_thread(self.other_category)

        response = self.client.post(
            self.api_link, {"new_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "You can't move posts to threads you can't reply."},
        )

    @patch_category_acl({"can_move_posts": True})
    def test_empty_data(self):
        """api handles empty data"""
        test.post_thread(self.category)

        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Enter link to new thread."})

    @patch_category_acl({"can_move_posts": True})
    def test_empty_posts_data_json(self):
        """api handles empty json data"""
        other_thread = test.post_thread(self.category)

        response = self.client.post(
            self.api_link,
            json.dumps({"new_thread": other_thread.get_absolute_url()}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "You have to specify at least one post to move."},
        )

    @patch_category_acl({"can_move_posts": True})
    def test_empty_posts_data_form(self):
        """api handles empty form data"""
        other_thread = test.post_thread(self.category)

        response = self.client.post(
            self.api_link, {"new_thread": other_thread.get_absolute_url()}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "You have to specify at least one post to move."},
        )

    @patch_category_acl({"can_move_posts": True})
    def test_no_posts_ids(self):
        """api rejects no posts ids"""
        other_thread = test.post_thread(self.category)

        response = self.client.post(
            self.api_link,
            json.dumps({"new_thread": other_thread.get_absolute_url(), "posts": []}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "You have to specify at least one post to move."},
        )

    @patch_category_acl({"can_move_posts": True})
    def test_invalid_posts_data(self):
        """api handles invalid data"""
        other_thread = test.post_thread(self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {"new_thread": other_thread.get_absolute_url(), "posts": "string"}
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": 'Expected a list of items but got type "str".'}
        )

    @patch_category_acl({"can_move_posts": True})
    def test_invalid_posts_ids(self):
        """api handles invalid post id"""
        other_thread = test.post_thread(self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "new_thread": other_thread.get_absolute_url(),
                    "posts": [1, 2, "string"],
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "One or more post ids received were invalid."}
        )

    @override_dynamic_settings(posts_per_page=5, posts_per_page_orphans=3)
    @patch_category_acl({"can_move_posts": True})
    def test_move_limit(self):
        """api rejects more posts than move limit"""
        other_thread = test.post_thread(self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {"new_thread": other_thread.get_absolute_url(), "posts": list(range(9))}
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "No more than 8 posts can be moved at a single time."},
        )

    @patch_category_acl({"can_move_posts": True})
    def test_move_invisible(self):
        """api validates posts visibility"""
        other_thread = test.post_thread(self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "new_thread": other_thread.get_absolute_url(),
                    "posts": [test.reply_thread(self.thread, is_unapproved=True).pk],
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "One or more posts to move could not be found."}
        )

    @patch_category_acl({"can_move_posts": True})
    def test_move_other_thread_posts(self):
        """api recjects attempt to move other thread's post"""
        other_thread = test.post_thread(self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "new_thread": other_thread.get_absolute_url(),
                    "posts": [test.reply_thread(other_thread, is_hidden=True).pk],
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "One or more posts to move could not be found."}
        )

    @patch_category_acl({"can_move_posts": True})
    def test_move_event(self):
        """api rejects events move"""
        other_thread = test.post_thread(self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "new_thread": other_thread.get_absolute_url(),
                    "posts": [test.reply_thread(self.thread, is_event=True).pk],
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Events can't be moved."})

    @patch_category_acl({"can_move_posts": True})
    def test_move_first_post(self):
        """api rejects first post move"""
        other_thread = test.post_thread(self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "new_thread": other_thread.get_absolute_url(),
                    "posts": [self.thread.first_post_id],
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "You can't move thread's first post."}
        )

    @patch_category_acl({"can_move_posts": True})
    def test_move_hidden_posts(self):
        """api recjects attempt to move urneadable hidden post"""
        other_thread = test.post_thread(self.category)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "new_thread": other_thread.get_absolute_url(),
                    "posts": [test.reply_thread(self.thread, is_hidden=True).pk],
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "You can't move posts the content you can't see."},
        )

    @patch_category_acl({"can_move_posts": True, "can_close_threads": False})
    def test_move_posts_closed_thread_no_permission(self):
        """api recjects attempt to move posts from closed thread"""
        other_thread = test.post_thread(self.category)

        self.thread.is_closed = True
        self.thread.save()

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "new_thread": other_thread.get_absolute_url(),
                    "posts": [test.reply_thread(self.thread).pk],
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "This thread is closed. You can't move posts in it."},
        )

    @patch_other_category_acl({"can_reply_threads": True, "can_close_threads": False})
    @patch_category_acl({"can_move_posts": True})
    def test_move_posts_closed_category_no_permission(self):
        """api recjects attempt to move posts from closed thread"""
        other_thread = test.post_thread(self.other_category)

        self.category.is_closed = True
        self.category.save()

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "new_thread": other_thread.get_absolute_url(),
                    "posts": [test.reply_thread(self.thread).pk],
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "This category is closed. You can't move posts in it."},
        )

    @patch_other_category_acl({"can_reply_threads": True})
    @patch_category_acl({"can_move_posts": True})
    def test_move_posts(self):
        """api moves posts to other thread"""
        other_thread = test.post_thread(self.other_category)

        posts = (
            test.reply_thread(self.thread).pk,
            test.reply_thread(self.thread).pk,
            test.reply_thread(self.thread).pk,
            test.reply_thread(self.thread).pk,
        )

        self.thread.refresh_from_db()
        self.assertEqual(self.thread.replies, 4)

        response = self.client.post(
            self.api_link,
            json.dumps({"new_thread": other_thread.get_absolute_url(), "posts": posts}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        # replies were moved
        self.thread.refresh_from_db()
        self.assertEqual(self.thread.replies, 0)

        other_thread = Thread.objects.get(pk=other_thread.pk)
        self.assertEqual(other_thread.post_set.filter(pk__in=posts).count(), 4)
        self.assertEqual(other_thread.replies, 4)

    @patch_other_category_acl({"can_reply_threads": True})
    @patch_category_acl({"can_move_posts": True})
    def test_move_best_answer(self):
        """api moves best answer to other thread"""
        other_thread = test.post_thread(self.other_category)
        best_answer = test.reply_thread(self.thread)

        self.thread.set_best_answer(self.user, best_answer)
        self.thread.synchronize()
        self.thread.save()

        self.thread.refresh_from_db()
        self.assertEqual(self.thread.best_answer, best_answer)
        self.assertEqual(self.thread.replies, 1)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "new_thread": other_thread.get_absolute_url(),
                    "posts": [best_answer.pk],
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        # best_answer was moved and unmarked
        self.thread.refresh_from_db()
        self.assertEqual(self.thread.replies, 0)
        self.assertIsNone(self.thread.best_answer)

        other_thread = Thread.objects.get(pk=other_thread.pk)
        self.assertEqual(other_thread.replies, 1)
        self.assertIsNone(other_thread.best_answer)

    @patch_other_category_acl({"can_reply_threads": True})
    @patch_category_acl({"can_move_posts": True})
    def test_move_posts_reads(self):
        """api moves posts reads together with posts"""
        other_thread = test.post_thread(self.other_category)

        posts = (test.reply_thread(self.thread), test.reply_thread(self.thread))

        self.thread.refresh_from_db()
        self.assertEqual(self.thread.replies, 2)

        poststracker.save_read(self.user, self.thread.first_post)
        for post in posts:
            poststracker.save_read(self.user, post)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "new_thread": other_thread.get_absolute_url(),
                    "posts": [p.pk for p in posts],
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        other_thread = Thread.objects.get(pk=other_thread.pk)

        # postreads were removed
        postreads = self.user.postread_set.order_by("id")

        postreads_threads = list(postreads.values_list("thread_id", flat=True))
        self.assertEqual(postreads_threads, [self.thread.pk])

        postreads_categories = list(postreads.values_list("category_id", flat=True))
        self.assertEqual(postreads_categories, [self.category.pk])



import pytest

from ..pluginlist import parse_plugins_list


def test_parser_raises_value_error_if_local_plugin_is_missing_path():
    with pytest.raises(ValueError):
        parse_plugins_list("plugin@")


def test_parser_raises_value_error_if_local_plugin_is_missing_module_name():
    with pytest.raises(ValueError):
        parse_plugins_list("@/local/")


def test_parser_raises_value_error_if_plugin_is_repeated():
    with pytest.raises(ValueError):
        parse_plugins_list("plugin\nplugin")


def test_parser_raises_value_error_if_local_plugin_is_repeated():
    with pytest.raises(ValueError):
        parse_plugins_list("plugin@/local/\n@plugin/other/local/")


def test_parser_raises_value_error_if_local_plugin_module_conflicts_with_other_plugin():
    with pytest.raises(ValueError):
        parse_plugins_list("plugin\nplugin@/local/")



# Generated by Django 1.9.7 on 2016-10-04 21:41
from django.db import migrations

ATTACHMENTS = [
    {
        "name": "GIF",
        "extensions": ("gif",),
        "mimetypes": ("image/gif",),
        "size_limit": 5 * 1024,
    },
    {
        "name": "JPG",
        "extensions": ("jpg", "jpeg"),
        "mimetypes": ("image/jpeg",),
        "size_limit": 3 * 1024,
    },
    {
        "name": "PNG",
        "extensions": ("png",),
        "mimetypes": ("image/png",),
        "size_limit": 3 * 1024,
    },
    {
        "name": "WEBP",
        "extensions": ("webp",),
        "mimetypes": ("image/webp",),
        "size_limit": 3 * 1024,
    },
    {
        "name": "PDF",
        "extensions": ("pdf",),
        "mimetypes": (
            "application/pdf",
            "application/x-pdf",
            "application/x-bzpdf",
            "application/x-gzpdf",
        ),
        "size_limit": 4 * 1024,
    },
    {
        "name": "Text",
        "extensions": ("txt",),
        "mimetypes": ("text/plain",),
        "size_limit": 4 * 1024,
    },
    {
        "name": "Markdown",
        "extensions": ("md",),
        "mimetypes": ("text/markdown",),
        "size_limit": 4 * 1024,
    },
    {
        "name": "reStructuredText",
        "extensions": ("rst",),
        "mimetypes": ("text/x-rst",),
        "size_limit": 4 * 1024,
    },
    {
        "name": "7Z",
        "extensions": ("7z",),
        "mimetypes": ("application/x-7z-compressed",),
        "size_limit": 4 * 1024,
    },
    {
        "name": "RAR",
        "extensions": ("rar",),
        "mimetypes": ("application/vnd.rar",),
        "size_limit": 4 * 1024,
    },
    {
        "name": "TAR",
        "extensions": ("tar",),
        "mimetypes": ("application/x-tar",),
        "size_limit": 4 * 1024,
    },
    {
        "name": "GZ",
        "extensions": ("gz",),
        "mimetypes": ("application/gzip",),
        "size_limit": 4 * 1024,
    },
    {
        "name": "ZIP",
        "extensions": ("zip", "zipx"),
        "mimetypes": ("application/zip",),
        "size_limit": 4 * 1024,
    },
]


def create_attachment_types(apps, schema_editor):
    AttachmentType = apps.get_model("misago_threads", "AttachmentType")
    for attachment in ATTACHMENTS:
        kwargs = attachment
        kwargs["extensions"] = ",".join(kwargs["extensions"])
        kwargs["mimetypes"] = ",".join(kwargs["mimetypes"])
        AttachmentType.objects.create(**kwargs)


class Migration(migrations.Migration):

    dependencies = [("misago_threads", "0002_threads_settings")]

    operations = [migrations.RunPython(create_attachment_types)]



# Generated by Django 1.9.7 on 2016-07-17 02:05
from django.db import migrations, models

import misago.users.models.user


class Migration(migrations.Migration):

    dependencies = [("misago_users", "0004_default_ranks")]

    operations = [
        migrations.AlterModelManagers(
            name="user", managers=[("objects", misago.users.models.user.UserManager())]
        ),
        migrations.AlterField(
            model_name="rank",
            name="roles",
            field=models.ManyToManyField(blank=True, to="misago_acl.Role"),
        ),
        migrations.AlterField(
            model_name="user",
            name="groups",
            field=models.ManyToManyField(
                blank=True,
                help_text="The groups this user belongs to. A user will get all permissions granted to each of their groups.",
                related_name="user_set",
                related_query_name="user",
                to="auth.Group",
                verbose_name="groups",
            ),
        ),
    ]



from django.db import models


class Icon(models.Model):
    TYPE_FAVICON = "favicon"
    TYPE_FAVICON_32 = "favicon_32"
    TYPE_FAVICON_16 = "favicon_16"

    TYPE_APPLE_TOUCH_ICON = "apple_touch_icon"

    FAVICON_TYPES = (TYPE_FAVICON, TYPE_FAVICON_32, TYPE_FAVICON_16)

    type = models.CharField(max_length=255)
    image = models.ImageField(
        upload_to="icon", height_field="height", width_field="width"
    )
    size = models.PositiveIntegerField(default=0)
    width = models.PositiveIntegerField(default=0)
    height = models.PositiveIntegerField(default=0)

    def delete(self, *args, **kwargs):
        if self.image:
            self.image.delete(save=False)
        return super().delete(*args, **kwargs)



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots["test_bold_bbcode 1"] = "<p>Lorem <b>ipsum</b>!</p>"

snapshots[
    "test_image_bbcode 1"
] = '<p>Lorem <img alt="placekitten.com/g/1200/500" src="https://placekitten.com/g/1200/500" /> ipsum</p>'

snapshots[
    "test_image_bbcode_is_case_insensitive 1"
] = '<p>Lorem <img alt="placekitten.com/g/1200/500" src="https://placekitten.com/g/1200/500" /> ipsum</p>'

snapshots[
    "test_image_bbcode_is_escaped 1"
] = '<p>Lorem <img alt="&lt;script language=&quot;application/javascript&quot;&gt;" src="http://&lt;script language=&quot;application/javascript&quot;&gt;" /> ipsum</p>'

snapshots["test_inline_bbcode_can_be_mixed 1"] = "<p>Lorem <b><u>ipsum</u></b>!</p>"

snapshots[
    "test_inline_bbcode_can_be_mixed_with_markdown 1"
] = "<p>Lorem <b><strong>ipsum</strong></b>!</p>"

snapshots["test_italics_bbcode 1"] = "<p>Lorem <i>ipsum</i>!</p>"

snapshots[
    "test_simple_inline_bbcode_is_escaped 1"
] = "<p>Lorem <b>ips &lt;script language=&quot;application/javascript&quot;&gt; um</b>!</p>"

snapshots["test_underline_bbcode 1"] = "<p>Lorem <u>ipsum</u>!</p>"

snapshots[
    "test_url_bbcode 1"
] = '<p>Lorem <a href="https://placekitten.com/g/1200/500" rel="external nofollow noopener" target="_blank">placekitten.com/g/1200/500</a> ipsum</p>'

snapshots[
    "test_url_bbcode_is_escaped 1"
] = '<p>Lorem <a href="http://&lt;script language=&quot;application/javascript&quot;&gt;" rel="external nofollow noopener" target="_blank">&lt;script language=&quot;application/javascript&quot;&gt;</a> ipsum</p>'

snapshots[
    "test_url_bbcode_link_text_is_escaped 1"
] = '<p>Lorem <a href="http://&lt;script language=&quot;application/javascript&quot;&gt;" rel="external nofollow noopener" target="_blank">&lt;script language=&quot;application/javascript&quot;&gt;</a> ipsum</p>'

snapshots[
    "test_url_bbcode_with_link_text 1"
] = '<p>Lorem <a href="https://placekitten.com/g/1200/500" rel="external nofollow noopener" target="_blank">dolor</a> ipsum</p>'

snapshots[
    "test_url_bbcode_with_long_link_text 1"
] = '<p>Lorem <a href="https://placekitten.com/g/1200/500" rel="external nofollow noopener" target="_blank">dolor met</a> ipsum</p>'

snapshots[
    "test_url_bbcode_with_quotes_and_link_text 1"
] = '<p>Lorem <a href="https://placekitten.com/g/1200/500" rel="external nofollow noopener" target="_blank">dolor</a> ipsum</p>'

snapshots[
    "test_url_bbcode_with_quotes_and_long_link_text 1"
] = '<p>Lorem <a href="https://placekitten.com/g/1200/500" rel="external nofollow noopener" target="_blank">dolor met</a> ipsum</p>'



from django import forms
from django.utils.translation import gettext as _

from ..models import AttachmentType


def get_searchable_filetypes():
    choices = [(0, _("All types"))]
    choices += [(a.id, a.name) for a in AttachmentType.objects.order_by("name")]
    return choices


class FilterAttachmentsForm(forms.Form):
    uploader = forms.CharField(label=_("Uploader name contains"), required=False)
    filename = forms.CharField(label=_("Filename contains"), required=False)
    filetype = forms.TypedChoiceField(
        label=_("File type"),
        coerce=int,
        choices=get_searchable_filetypes,
        empty_value=0,
        required=False,
    )
    is_orphan = forms.ChoiceField(
        label=_("State"),
        required=False,
        choices=[
            ("", _("All")),
            ("yes", _("Only orphaned")),
            ("no", _("Not orphaned")),
        ],
    )

    def filter_queryset(self, criteria, queryset):
        if criteria.get("uploader"):
            queryset = queryset.filter(
                uploader_slug__contains=criteria["uploader"].lower()
            )
        if criteria.get("filename"):
            queryset = queryset.filter(filename__icontains=criteria["filename"])
        if criteria.get("filetype"):
            queryset = queryset.filter(filetype_id=criteria["filetype"])
        if criteria.get("is_orphan") == "yes":
            queryset = queryset.filter(post__isnull=True)
        elif criteria.get("is_orphan") == "no":
            queryset = queryset.filter(post__isnull=False)
        return queryset


class AttachmentTypeForm(forms.ModelForm):
    class Meta:
        model = AttachmentType
        fields = "__all__"
        labels = {
            "name": _("Type name"),
            "extensions": _("File extensions"),
            "mimetypes": _("Mimetypes"),
            "size_limit": _("Maximum allowed uploaded file size"),
            "status": _("Status"),
            "limit_uploads_to": _("Limit uploads to"),
            "limit_downloads_to": _("Limit downloads to"),
        }
        help_texts = {
            "extensions": _(
                "List of comma separated file extensions associated with this "
                "attachment type."
            ),
            "mimetypes": _(
                "Optional list of comma separated mime types associated with this "
                "attachment type."
            ),
            "size_limit": _(
                "Maximum allowed uploaded file size for this type, in kb. "
                "This setting is deprecated and has no effect. It will be "
                "deleted in Misago 1.0."
            ),
            "status": _("Controls this attachment type availability on your site."),
            "limit_uploads_to": _(
                "If you wish to limit option to upload files of this type to users "
                "with specific roles, select them on this list. Otherwhise don't "
                "select any roles to allow all users with permission to upload "
                "attachments to be able to upload attachments of this type."
            ),
            "limit_downloads_to": _(
                "If you wish to limit option to download files of this type to users "
                "with specific roles, select them on this list. Otherwhise don't "
                "select any roles to allow all users with permission to download "
                "attachments to be able to download attachments of this type."
            ),
        }
        widgets = {
            "limit_uploads_to": forms.CheckboxSelectMultiple,
            "limit_downloads_to": forms.CheckboxSelectMultiple,
        }

    def clean_extensions(self):
        data = self.clean_list(self.cleaned_data["extensions"])
        if not data:
            raise forms.ValidationError(_("This field is required."))
        return data

    def clean_mimetypes(self):
        data = self.cleaned_data["mimetypes"]
        if data:
            return self.clean_list(data)

    def clean_list(self, value):
        items = [v.lstrip(".") for v in value.lower().replace(" ", "").split(",")]
        return ",".join(set(filter(bool, items)))



import os

import pytest
from django.core.files.uploadedfile import UploadedFile
from django.urls import reverse

from ....test import assert_has_error_message

TESTS_DIR = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def png_file():
    return os.path.join(TESTS_DIR, "images", "test.png")


@pytest.fixture
def svg_file():
    return os.path.join(TESTS_DIR, "images", "test.svg")


@pytest.fixture
def hashable_file():
    return os.path.join(TESTS_DIR, "css", "test.css")


@pytest.fixture
def hashed_file():
    return os.path.join(TESTS_DIR, "css", "test.4846cb3b.css")


@pytest.fixture
def upload(admin_client):
    def post_upload(theme, asset_files=None):
        url = reverse("misago:admin:themes:upload-media", kwargs={"pk": theme.pk})
        if asset_files is not None:
            data = asset_files if isinstance(asset_files, list) else [asset_files]
        else:
            data = ""
        return admin_client.post(url, {"assets": data})

    return post_upload


def test_font_file_can_be_uploaded(upload, theme):
    font_file = os.path.join(TESTS_DIR, "font", "Lato.ttf")
    with open(font_file, "rb") as fp:
        upload(theme, fp)
        assert theme.media.exists()


def test_text_file_can_be_uploaded(upload, theme):
    text_file = os.path.join(TESTS_DIR, "font", "OFL.txt")
    with open(text_file) as fp:
        upload(theme, fp)
        assert theme.media.exists()


def test_svg_file_can_be_uploaded(upload, theme):
    svg_file = os.path.join(TESTS_DIR, "images", "test.svg")
    with open(svg_file) as fp:
        upload(theme, fp)
        assert theme.media.exists()


def test_png_file_can_be_uploaded(upload, theme, png_file):
    with open(png_file, "rb") as fp:
        upload(theme, fp)
        assert theme.media.exists()


def test_media_file_name_is_set_as_asset_name(upload, theme, svg_file):
    with open(svg_file) as fp:
        upload(theme, fp)

    media = theme.media.last()
    expected_filename = str(svg_file).split("/")[-1]
    assert media.name == expected_filename


def test_media_file_is_uploaded_to_theme_directory(upload, theme, svg_file):
    with open(svg_file) as fp:
        upload(theme, fp)

    media = theme.media.last()
    assert theme.dirname in str(media.file)


def test_hash_is_added_to_uploaded_media_file_name(
    upload, theme, hashable_file, hashed_file
):
    with open(hashable_file) as fp:
        upload(theme, fp)

    media = theme.media.last()
    filename = str(media.file.path).split("/")[-1]
    expected_filename = str(hashed_file).split("/")[-1]
    assert filename == expected_filename


def test_hash_is_set_on_media_asset(upload, theme, hashed_file):
    with open(hashed_file) as fp:
        upload(theme, fp)

    media = theme.media.last()
    assert media.hash


def test_media_file_name_is_preserved_if_it_already_contains_correct_hash(
    upload, theme, hashed_file
):
    with open(hashed_file) as fp:
        upload(theme, fp)

    media = theme.media.last()
    filename = str(media.file.path).split("/")[-1]
    expected_filename = str(hashed_file).split("/")[-1]
    assert filename == expected_filename


def test_new_hash_is_added_to_media_file_name_if_it_contains_incorrect_hash(
    upload, theme
):
    incorrectly_hashed_file = os.path.join(TESTS_DIR, "css", "test.0046cb3b.css")
    with open(incorrectly_hashed_file) as fp:
        upload(theme, fp)

    media = theme.media.last()
    filename = str(media.file.path).split("/")[-1]
    assert media.hash in filename


def test_newly_uploaded_media_file_replaces_old_one_if_file_names_are_same(
    upload, theme, hashable_file
):
    with open(hashable_file) as fp:
        upload(theme, fp)
    original_media = theme.media.get()

    with open(os.path.join(TESTS_DIR, "css", "test-changed.css")) as fp:
        size = len(fp.read())
        fp.seek(0)
        upload(
            theme, UploadedFile(fp, name="test.css", content_type="text/css", size=size)
        )
    updated_media = theme.media.last()

    assert updated_media.name == original_media.name
    assert updated_media.hash != original_media.hash
    assert theme.media.count() == 1


def test_image_dimensions_are_set_for_uploaded_image_file(upload, theme, png_file):
    with open(png_file, "rb") as fp:
        upload(theme, fp)

    media = theme.media.last()
    assert media.width
    assert media.height


def test_thumbnail_is_generated_in_theme_directory_for_uploaded_image_file(
    upload, theme, png_file
):
    with open(png_file, "rb") as fp:
        upload(theme, fp)

    media = theme.media.last()
    assert theme.dirname in str(media.thumbnail)


def test_uploading_media_triggers_css_build(
    upload, theme, png_file, mock_build_theme_css
):
    with open(png_file, "rb") as fp:
        upload(theme, fp)

    mock_build_theme_css.assert_called_once_with(theme.pk)


def test_error_message_is_set_if_no_media_was_uploaded(upload, theme):
    response = upload(theme)
    assert_has_error_message(response)


def test_error_message_is_set_if_user_attempts_to_upload_file_to_default_theme(
    upload, default_theme
):
    response = upload(default_theme)
    assert_has_error_message(response)


def test_error_message_is_set_if_user_attempts_to_upload_file_to_nonexisting_theme(
    upload, nonexisting_theme
):
    response = upload(nonexisting_theme)
    assert_has_error_message(response)



from ariadne_django.views import GraphQLView
from django.urls import path

from .schema import schema


class MisagoAdminExtension:
    def register_urlpatterns(self, urlpatterns):
        # GraphQL API
        urlpatterns.namespace("graphql/", "graphql")
        urlpatterns.patterns(
            "graphql", path("", GraphQLView.as_view(schema=schema), name="index")
        )



from django.test.client import BOUNDARY, MULTIPART_CONTENT, encode_multipart
from django.urls import reverse

from .. import test
from ...categories.models import Category
from ...markup.mentions import MENTIONS_LIMIT
from ...users.test import AuthenticatedUserTestCase, create_test_user


class PostMentionsTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")
        self.thread = test.post_thread(category=self.category)

        self.post_link = reverse(
            "misago:api:thread-post-list", kwargs={"thread_pk": self.thread.pk}
        )

    def put(self, url, data=None):
        content = encode_multipart(BOUNDARY, data or {})
        return self.client.put(url, content, content_type=MULTIPART_CONTENT)

    def test_mention_noone(self):
        """endpoint handles no mentions in post"""
        response = self.client.post(
            self.post_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        post = self.user.post_set.order_by("id").last()
        self.assertEqual(post.mentions.count(), 0)

    def test_mention_nonexistant(self):
        """endpoint handles nonexistant mention"""
        response = self.client.post(
            self.post_link, data={"post": "This is test response, @InvalidUser!"}
        )
        self.assertEqual(response.status_code, 200)

        post = self.user.post_set.order_by("id").last()
        self.assertEqual(post.mentions.count(), 0)

    def test_mention_self(self):
        """endpoint mentions author"""
        response = self.client.post(
            self.post_link, data={"post": "This is test response, @%s!" % self.user}
        )
        self.assertEqual(response.status_code, 200)

        post = self.user.post_set.order_by("id").last()

        self.assertEqual(post.mentions.count(), 1)
        self.assertEqual(post.mentions.all()[0], self.user)

    def test_mention_limit(self):
        """endpoint mentions over limit results in no mentions set"""
        users = []

        for i in range(MENTIONS_LIMIT + 5):
            users.append(create_test_user("User%s" % i, "user%s@example.com" % i))

        mentions = ["@%s" % u for u in users]
        response = self.client.post(
            self.post_link,
            data={"post": "This is test response, %s!" % (", ".join(mentions))},
        )
        self.assertEqual(response.status_code, 200)

        post = self.user.post_set.order_by("id").last()

        self.assertEqual(post.mentions.count(), 0)

    def test_mention_update(self):
        """edit post endpoint updates mentions"""
        user = create_test_user("User", "user@example.com")
        other_user = create_test_user("OtherUser", "otheruser@example.com")

        response = self.client.post(
            self.post_link, data={"post": "This is test response, @%s!" % user}
        )
        self.assertEqual(response.status_code, 200)

        post = self.user.post_set.order_by("id").last()

        self.assertEqual(post.mentions.count(), 1)
        self.assertEqual(post.mentions.order_by("id")[0], user)

        # add mention to post
        edit_link = reverse(
            "misago:api:thread-post-detail",
            kwargs={"thread_pk": self.thread.pk, "pk": post.pk},
        )

        response = self.put(
            edit_link,
            data={"post": "This is test response, @%s and @%s!" % (user, other_user)},
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(post.mentions.count(), 2)
        self.assertEqual(list(post.mentions.order_by("id")), [user, other_user])

        # remove first mention from post - should preserve mentions
        response = self.put(
            edit_link, data={"post": "This is test response, @%s!" % other_user}
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(post.mentions.count(), 2)
        self.assertEqual(list(post.mentions.order_by("id")), [user, other_user])

        # remove mentions from post - should preserve mentions
        response = self.put(edit_link, data={"post": "This is test response!"})
        self.assertEqual(response.status_code, 200)

        self.assertEqual(post.mentions.count(), 2)
        self.assertEqual(list(post.mentions.order_by("id")), [user, other_user])

    def test_mentions_merge(self):
        """posts merge sums mentions"""
        user = create_test_user("User1", "user1@example.com")
        other_user = create_test_user("User2", "user2@example.com")

        response = self.client.post(
            self.post_link, data={"post": "This is test response, @%s!" % user}
        )
        self.assertEqual(response.status_code, 200)

        post_a = self.user.post_set.order_by("id").last()

        self.assertEqual(post_a.mentions.count(), 1)
        self.assertEqual(list(post_a.mentions.all()), [user])

        # post second reply
        self.user.last_post_on = None
        self.user.save()

        response = self.client.post(
            self.post_link,
            data={"post": "This is test response, @%s and @%s!" % (user, other_user)},
        )
        self.assertEqual(response.status_code, 200)

        post_b = self.user.post_set.order_by("id").last()

        # merge posts and validate that post A has all mentions
        post_b.merge(post_a)

        self.assertEqual(post_a.mentions.count(), 2)
        self.assertEqual(list(post_a.mentions.order_by("id")), [user, other_user])



from django import forms
from django.core.exceptions import PermissionDenied
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

from ...acl import algebra
from ...acl.decorators import return_boolean
from ...categories.models import Category, CategoryRole
from ...categories.permissions import get_categories_roles
from ..models import Post, Thread

__all__nope = [
    "allow_mark_best_answer",
    "can_mark_best_answer",
    "allow_mark_as_best_answer",
    "can_mark_as_best_answer",
    "allow_unmark_best_answer",
    "can_unmark_best_answer",
    "allow_hide_best_answer",
    "can_hide_best_answer",
    "allow_delete_best_answer",
    "can_delete_best_answer",
]


class CategoryPermissionsForm(forms.Form):
    legend = _("Best answers")

    can_mark_best_answers = forms.TypedChoiceField(
        label=_("Can mark posts as best answers"),
        coerce=int,
        initial=0,
        choices=[(0, _("No")), (1, _("Own threads")), (2, _("All threads"))],
    )
    can_change_marked_answers = forms.TypedChoiceField(
        label=_("Can change marked answers"),
        coerce=int,
        initial=0,
        choices=[(0, _("No")), (1, _("Own threads")), (2, _("All threads"))],
    )
    best_answer_change_time = forms.IntegerField(
        label=_(
            "Time limit for changing marked best answer in owned thread, in minutes"
        ),
        help_text=_(
            "Enter 0 to don't limit time for changing marked best answer in "
            "owned thread."
        ),
        initial=0,
        min_value=0,
    )


def change_permissions_form(role):
    if isinstance(role, CategoryRole):
        return CategoryPermissionsForm


def build_acl(acl, roles, key_name):
    categories_roles = get_categories_roles(roles)
    categories = list(Category.objects.all_categories(include_root=True))

    for category in categories:
        category_acl = acl["categories"].get(category.pk, {"can_browse": 0})
        if category_acl["can_browse"]:
            acl["categories"][category.pk] = build_category_acl(
                category_acl, category, categories_roles, key_name
            )

    private_category = Category.objects.private_threads()
    private_threads_acl = acl["categories"].get(private_category.pk)
    if private_threads_acl:
        private_threads_acl.update(
            {
                "can_mark_best_answers": 0,
                "can_change_marked_answers": 0,
                "best_answer_change_time": 0,
            }
        )

    return acl


def build_category_acl(acl, category, categories_roles, key_name):
    category_roles = categories_roles.get(category.pk, [])

    final_acl = {
        "can_mark_best_answers": 0,
        "can_change_marked_answers": 0,
        "best_answer_change_time": 0,
    }
    final_acl.update(acl)

    algebra.sum_acls(
        final_acl,
        roles=category_roles,
        key=key_name,
        can_mark_best_answers=algebra.greater,
        can_change_marked_answers=algebra.greater,
        best_answer_change_time=algebra.greater_or_zero,
    )

    return final_acl


def add_acl_to_thread(user_acl, thread):
    thread.acl.update(
        {
            "can_mark_best_answer": can_mark_best_answer(user_acl, thread),
            "can_change_best_answer": can_change_best_answer(user_acl, thread),
            "can_unmark_best_answer": can_unmark_best_answer(user_acl, thread),
        }
    )


def add_acl_to_post(user_acl, post):
    post.acl.update(
        {
            "can_mark_as_best_answer": can_mark_as_best_answer(user_acl, post),
            "can_hide_best_answer": can_hide_best_answer(user_acl, post),
            "can_delete_best_answer": can_delete_best_answer(user_acl, post),
        }
    )


def register_with(registry):
    registry.acl_annotator(Thread, add_acl_to_thread)
    registry.acl_annotator(Post, add_acl_to_post)


def allow_mark_best_answer(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(_("You have to sign in to mark best answers."))

    category_acl = user_acl["categories"].get(target.category_id, {})

    if not category_acl.get("can_mark_best_answers"):
        raise PermissionDenied(
            _(
                "You don't have permission to mark best answers in the "
                '"%(category)s" category.'
            )
            % {"category": target.category}
        )

    if (
        category_acl["can_mark_best_answers"] == 1
        and user_acl["user_id"] != target.starter_id
    ):
        raise PermissionDenied(
            _(
                "You don't have permission to mark best answer in this thread "
                "because you didn't start it."
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                _(
                    "You don't have permission to mark best answer in this thread "
                    'because its category "%(category)s" is closed.'
                )
                % {"category": target.category}
            )
        if target.is_closed:
            raise PermissionDenied(
                _(
                    "You can't mark best answer in this thread because it's closed "
                    "and you don't have permission to open it."
                )
            )


can_mark_best_answer = return_boolean(allow_mark_best_answer)


def allow_change_best_answer(user_acl, target):
    if not target.has_best_answer:
        return  # shortcircut permission test

    category_acl = user_acl["categories"].get(target.category_id, {})

    if not category_acl.get("can_change_marked_answers"):
        raise PermissionDenied(
            _(
                "You don't have permission to change this thread's marked answer "
                'because it\'s in the "%(category)s" category.'
            )
            % {"category": target.category}
        )

    if category_acl["can_change_marked_answers"] == 1:
        if user_acl["user_id"] != target.starter_id:
            raise PermissionDenied(
                _(
                    "You don't have permission to change this thread's marked answer "
                    "because you are not a thread starter."
                )
            )
        if not has_time_to_change_answer(user_acl, target):
            # pylint: disable=line-too-long
            message = ngettext(
                "You don't have permission to change best answer that was marked for more than %(minutes)s minute.",
                "You don't have permission to change best answer that was marked for more than %(minutes)s minutes.",
                category_acl["best_answer_change_time"],
            )
            raise PermissionDenied(
                message % {"minutes": category_acl["best_answer_change_time"]}
            )

    if target.best_answer_is_protected and not category_acl["can_protect_posts"]:
        raise PermissionDenied(
            _(
                "You don't have permission to change this thread's best answer "
                "because a moderator has protected it."
            )
        )


can_change_best_answer = return_boolean(allow_change_best_answer)


def allow_unmark_best_answer(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(_("You have to sign in to unmark best answers."))

    if not target.has_best_answer:
        return  # shortcircut test

    category_acl = user_acl["categories"].get(target.category_id, {})

    if not category_acl.get("can_change_marked_answers"):
        raise PermissionDenied(
            _(
                "You don't have permission to unmark threads answers in "
                'the "%(category)s" category.'
            )
            % {"category": target.category}
        )

    if category_acl["can_change_marked_answers"] == 1:
        if user_acl["user_id"] != target.starter_id:
            raise PermissionDenied(
                _(
                    "You don't have permission to unmark this best answer "
                    "because you are not a thread starter."
                )
            )
        if not has_time_to_change_answer(user_acl, target):
            # pylint: disable=line-too-long
            message = ngettext(
                "You don't have permission to unmark best answer that was marked for more than %(minutes)s minute.",
                "You don't have permission to unmark best answer that was marked for more than %(minutes)s minutes.",
                category_acl["best_answer_change_time"],
            )
            raise PermissionDenied(
                message % {"minutes": category_acl["best_answer_change_time"]}
            )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                _(
                    "You don't have permission to unmark this best answer "
                    'because its category "%(category)s" is closed.'
                )
                % {"category": target.category}
            )
        if target.is_closed:
            raise PermissionDenied(
                _(
                    "You can't unmark this thread's best answer "
                    "because it's closed and you don't have permission to open it."
                )
            )

    if target.best_answer_is_protected and not category_acl["can_protect_posts"]:
        raise PermissionDenied(
            _(
                "You don't have permission to unmark this thread's best answer "
                "because a moderator has protected it."
            )
        )


can_unmark_best_answer = return_boolean(allow_unmark_best_answer)


def allow_mark_as_best_answer(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(_("You have to sign in to mark best answers."))

    if target.is_event:
        raise PermissionDenied(_("Events can't be marked as best answers."))

    category_acl = user_acl["categories"].get(target.category_id, {})

    if not category_acl.get("can_mark_best_answers"):
        raise PermissionDenied(
            _(
                "You don't have permission to mark best answers "
                'in the "%(category)s" category.'
            )
            % {"category": target.category}
        )

    if (
        category_acl["can_mark_best_answers"] == 1
        and user_acl["user_id"] != target.thread.starter_id
    ):
        raise PermissionDenied(
            _(
                "You don't have permission to mark best answer in this thread "
                "because you didn't start it."
            )
        )

    if target.is_first_post:
        raise PermissionDenied(
            _("First post in a thread can't be marked as best answer.")
        )

    if target.is_hidden:
        raise PermissionDenied(_("Hidden posts can't be marked as best answers."))

    if target.is_unapproved:
        raise PermissionDenied(_("Unapproved posts can't be marked as best answers."))

    if target.is_protected and not category_acl["can_protect_posts"]:
        raise PermissionDenied(
            _(
                "You don't have permission to mark this post as best answer "
                "because a moderator has protected it."
            )
        )


can_mark_as_best_answer = return_boolean(allow_mark_as_best_answer)


def allow_hide_best_answer(user_acl, target):
    if target.is_best_answer:
        raise PermissionDenied(
            _("You can't hide this post because its marked as best answer.")
        )


can_hide_best_answer = return_boolean(allow_hide_best_answer)


def allow_delete_best_answer(user_acl, target):
    if target.is_best_answer:
        raise PermissionDenied(
            _("You can't delete this post because its marked as best answer.")
        )


can_delete_best_answer = return_boolean(allow_delete_best_answer)


def has_time_to_change_answer(user_acl, target):
    category_acl = user_acl["categories"].get(target.category_id, {})
    change_time = category_acl.get("best_answer_change_time", 0)

    if change_time:
        diff = timezone.now() - target.best_answer_marked_on
        diff_minutes = int(diff.total_seconds() / 60)
        return diff_minutes < change_time

    return True



"""
Misago saves parsed strings in database.

Those strings are "trusted" and contain HTML that is rendered by templates
without additional sanitization step.

While this greatly improves speed, it also means that SQLInjections may
escalate to Code Injection vulnerabilities.

Because of this you should use this module to generate checksum for each model
that contains parsed strings. Each checksum should be generated from markup as
well as additional unique values that are specific for that model, like its PK,
post date, etc ect.

That way even if few items will contain the same content, they will have
different checksums, and as long as attacker has no access to filesystem,
he'll wont know SECRET_KEY and thus won't be able to generate valid checksums
for injected content

Because SHA256 is used for checksum generation, make sure you are storing them
in char fields with max_length=64
"""
from hashlib import sha256


def make_checksum(parsed, unique_values=None):
    unique_values = unique_values or []
    seeds = [parsed] + [str(v) for v in unique_values]

    return sha256("+".join(seeds).encode("utf-8")).hexdigest()


def is_checksum_valid(parsed, checksum, unique_values=None):
    return checksum == make_checksum(parsed, unique_values)



from unittest.mock import Mock

import pytest

from ..hydrators import dehydrate_value, hydrate_value


def test_string_value_can_be_dehydrated_and_hydrated_back():
    assert hydrate_value("string", dehydrate_value("string", "test")) == "test"


def test_int_value_is_dehydrated_to_string():
    assert dehydrate_value("string", 123) == "123"


def test_empty_string_value_is_hydrated_to_empty_string():
    assert hydrate_value("string", None) == ""


def test_bool_false_value_can_be_dehydrated_and_hydrated_back():
    assert hydrate_value("bool", dehydrate_value("bool", False)) is False


def test_bool_true_value_can_be_dehydrated_and_hydrated_back():
    assert hydrate_value("bool", dehydrate_value("bool", True)) is True


def test_bool_none_value_can_be_dehydrated_and_hydrated_back_to_false():
    assert hydrate_value("bool", dehydrate_value("bool", None)) is False


def test_int_value_can_be_dehydrated_and_hydrated_back():
    assert hydrate_value("int", dehydrate_value("int", 123)) == 123


def test_empty_int_value_can_be_dehydrated_and_hydrated_back():
    assert hydrate_value("int", dehydrate_value("int", 0)) == 0


def test_none_int_value_is_dehydrated_to_zero_string():
    assert dehydrate_value("int", None) == "0"


def test_none_int_value_is_hydrated_to_zero():
    assert hydrate_value("int", None) == 0


def test_empty_int_value_is_hydrated_to_zero():
    assert hydrate_value("int", "") == 0


def test_list_value_can_be_dehydrated_and_hydrated_back():
    assert hydrate_value("list", dehydrate_value("list", ["a", "b"])) == ["a", "b"]


def test_single_item_list_value_can_be_dehydrated_and_hydrated_back():
    assert hydrate_value("list", dehydrate_value("list", ["a"])) == ["a"]


def test_empty_list_value_can_be_dehydrated_and_hydrated_back():
    assert hydrate_value("list", dehydrate_value("list", [])) == []


def test_none_list_value_can_be_dehydrated_and_hydrated_to_empty_list():
    assert hydrate_value("list", dehydrate_value("list", None)) == []


def test_empty_list_value_is_hydrated_to_empty_list():
    assert hydrate_value("list", "") == []


def test_none_list_value_is_hydrated_to_empty_list():
    assert hydrate_value("list", None) == []


def test_none_list_value_is_dehydrated_to_empty_string():
    assert dehydrate_value("list", None) == ""


def test_image_value_hydration_is_noop():
    image = Mock()
    assert hydrate_value("image", image) is image


def test_image_value_dehydration_is_noop():
    image = Mock()
    assert dehydrate_value("image", image) is image


def test_value_error_is_raised_on_unsupported_type_dehydration():
    with pytest.raises(ValueError):
        dehydrate_value("unsupported", None)


def test_value_error_is_raised_on_unsupported_type_hydration():
    with pytest.raises(ValueError):
        hydrate_value("unsupported", None)



# Generated by Django 1.11.13 on 2018-06-09 15:23
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("misago_threads", "0009_auto_20180326_0010")]

    operations = [
        migrations.RemoveField(model_name="attachment", name="uploader_ip"),
        migrations.RemoveField(model_name="poll", name="poster_ip"),
        migrations.RemoveField(model_name="pollvote", name="voter_ip"),
        migrations.RemoveField(model_name="post", name="poster_ip"),
        migrations.RemoveField(model_name="postedit", name="editor_ip"),
        migrations.RemoveField(model_name="postlike", name="liker_ip"),
    ]
